"""Functions for thresholding and segmenting images."""

import matplotlib.pyplot as plt
import numba as nb
import numpy as np
import skfmm
from numba import njit, prange
from numpy.typing import NDArray
from scipy import ndimage
from skimage import filters
from skimage.feature import peak_local_max
from skimage.morphology import binary_erosion, disk
from skimage.morphology import dilation as sk_dilation
from skimage.segmentation import watershed
from skimage.transform import resize

from scportrait.plotting.vis import plot_image

DEFAULT_SEGMENTATION_DTYPE = np.uint32


def global_otsu(image: NDArray) -> float:
    """Calculate optimal global threshold using Otsu's method.

    Args:
        image: Input grayscale image

    Returns:
        Optimal threshold value calculated using Otsu's method

    Examples:
        >>> from skimage import data
        >>> image = data.coins()
        >>> threshold = global_otsu(image)
        >>> print(threshold)
    """
    counts, bin_edges = np.histogram(np.ravel(image), bins=512)
    bin_centers = bin_edges[:-1] + np.diff(bin_edges) / 2

    weight1 = np.cumsum(counts)
    weight2 = np.cumsum(counts[::-1])[::-1]

    mean1 = np.cumsum(counts * bin_centers) / weight1
    mean2 = (np.cumsum((counts * bin_centers)[::-1]) / weight2[::-1])[::-1]

    variance12 = weight1[:-1] * weight2[1:] * (mean1[:-1] - mean2[1:]) ** 2

    idx = np.argmax(variance12)
    threshold = bin_centers[idx]

    return threshold


def _segment_threshold(
    image: NDArray,
    threshold: float,
    speckle_kernel: int = 4,
    debug: bool = False,
) -> NDArray:
    """Segment image using threshold and additional parameters.

    Args:
        image: Input grayscale image
        threshold: Threshold value for creating image_mask
        speckle_kernel: Size of structuring element for speckle removal
        debug: Enable debug visualization

    Returns:
        Binary mask after cleaning
    """
    image_mask = image > threshold

    if debug:
        plot_image(image_mask, cmap="Greys_r")

    image_mask_clean = binary_erosion(image_mask, footprint=disk(speckle_kernel))
    image_mask_clean = sk_dilation(image_mask_clean, footprint=disk(speckle_kernel - 1))

    return image_mask_clean


def _generate_labels_from_mask(
    image_mask: NDArray, dilation: int = 4, min_distance: int = 10, peak_footprint: int = 7, debug: bool = False
) -> NDArray:
    """Generate labels from binary mask using distance transform and watershed.

    Args:
        image_mask: Binary mask to generate labels from
        dilation: Size of dilation structuring element
        min_distance: Minimum distance between peaks
        peak_footprint: Size of peak finding structuring element
        debug: Enable debug visualization

    Returns:
        Label array where each segment has unique label
    """
    distance = ndimage.distance_transform_edt(image_mask)
    peak_idx = peak_local_max(distance, min_distance=min_distance, footprint=disk(peak_footprint))
    local_maxi = np.zeros_like(image_mask, dtype=bool)
    local_maxi[tuple(peak_idx.T)] = True

    if debug:
        plt.figure(frameon=False, figsize=(10, 10))
        plt.imshow(image_mask, cmap="Greys_r")
        plt.title("Generated Mask")
        plt.axis("off")
        plt.scatter(peak_idx[:, 1], peak_idx[:, 0], s=2, color="red")
        plt.show()

    dilated_mask = sk_dilation(image_mask, footprint=disk(dilation))

    fmm_marker = np.ones_like(dilated_mask)
    for center in peak_idx:
        fmm_marker[center[0], center[1]] = 0

    m = np.ma.masked_array(fmm_marker, np.logical_not(dilated_mask))
    distance_2 = skfmm.distance(m)

    if debug:
        plt.figure(frameon=False, figsize=(10, 10))
        plt.title("Distance Transform")
        plt.imshow(distance_2, cmap="viridis")
        plt.scatter(peak_idx[:, 1], peak_idx[:, 0], s=2, color="red")
        plt.axis("off")
        plt.show()

    marker = np.zeros_like(image_mask).astype(DEFAULT_SEGMENTATION_DTYPE)
    for i, center in enumerate(peak_idx):
        marker[center[0], center[1]] = i + 1

    labels = watershed(distance_2, marker, mask=dilated_mask)
    return labels.astype(DEFAULT_SEGMENTATION_DTYPE)


def segment_global_threshold(
    image: NDArray,
    dilation: int = 4,
    min_distance: int = 10,
    peak_footprint: int = 7,
    speckle_kernel: int = 4,
    debug: bool = False,
) -> NDArray:
    """Segment image using Otsu's method and watershed.

    Args:
        image: Input grayscale image
        dilation: Size of dilation structuring element
        min_distance: Minimum distance between peaks
        peak_footprint: Size of peak finding structuring element
        speckle_kernel: Size of speckle removal structuring element
        debug: Enable debug visualization

    Returns:
        Label array where each segment has unique label

    Example:
        >>> from skimage import data
        >>> coins = data.coins()
        >>> labels = segment_global_threshold(coins, dilation=5, min_distance=20)
    """
    threshold = global_otsu(image)

    mask = _segment_threshold(
        image,
        threshold,
        speckle_kernel=speckle_kernel,
        debug=debug,
    )

    labels = _generate_labels_from_mask(
        mask, dilation=dilation, min_distance=min_distance, peak_footprint=peak_footprint, debug=debug
    )

    return labels


def segment_local_threshold(
    image: NDArray,
    dilation: int = 4,
    thr: float = 0.01,
    median_block: int = 51,
    min_distance: int = 10,
    peak_footprint: int = 7,
    speckle_kernel: int = 4,
    median_step: int = 1,
    debug: bool = False,
) -> NDArray:
    """Segment image using local threshold method.

    Args:
        image: Input grayscale image
        dilation: Size of dilation structuring element
        thr: Value added to block median for threshold
        median_block: Size of local threshold block
        min_distance: Minimum distance between peaks
        peak_footprint: Size of peak finding structuring element
        speckle_kernel: Size of speckle removal structuring element
        median_step: Step size for downsampling
        debug: Enable debug visualization

    Returns:
        Label array where each segment has unique label

    Raises:
        ValueError: If speckle_kernel < 1
    """
    if speckle_kernel < 1:
        raise ValueError("speckle_kernel needs to be at least 1")

    # Downsample the input image
    downsampled_image = image[::median_step, ::median_step]

    # Calculate local threshold
    local_threshold = filters.threshold_local(downsampled_image, block_size=median_block, method="median", offset=-thr)

    # Resize the local threshold image
    local_threshold = resize(local_threshold, image.shape)

    # Segment using local threshold
    mask = _segment_threshold(
        image,
        local_threshold,
        speckle_kernel=speckle_kernel,
        debug=debug,
    )

    labels = _generate_labels_from_mask(
        mask, dilation=dilation, min_distance=min_distance, peak_footprint=peak_footprint, debug=debug
    )

    return labels


@njit
def _numba_subtract(array1: NDArray, number: int) -> NDArray:
    """Subtract a number from all non-zero elements of array.

    Args:
        array1: Input array
        number: Number to subtract from non-zero elements

    Returns:
        Array with number subtracted from non-zero elements

    Example:
        >>> array1 = np.array([[0, 2, 3], [0, 5, 6], [0, 0, 7]])
        >>> _numba_subtract(array1, 1)
        array([[0, 1, 2],
               [0, 4, 5],
               [0, 0, 6]])
    """
    for i in range(array1.shape[0]):
        for j in range(array1.shape[1]):
            if array1[i, j] != 0:
                array1[i, j] = array1[i, j] - number
    return array1


@njit
def _return_edge_labels_2d(input_map: NDArray) -> list[int]:
    """Get labels touching edges in 2D array."""
    top_row = input_map[0]
    bottom_row = input_map[-1]
    first_column = input_map[:, 0]
    last_column = input_map[:, -1]

    full_union = (
        set(top_row.flatten())
        .union(set(bottom_row.flatten()))
        .union(set(first_column.flatten()))
        .union(set(last_column.flatten()))
    )

    full_union = set([np.uint64(i) for i in full_union])  # noqa: C403
    full_union.discard(0)

    return list(full_union)


def _return_edge_labels(input_map: NDArray) -> list[int]:
    """Get labels touching edges in 2D/3D array.

    Args:
        input_map: Input segmentation array (2D or 3D)

    Returns:
        List of unique labels touching edges
    """
    if len(input_map.shape) == 3:
        full_union: list[int] = []
        for i in range(input_map.shape[0]):
            _ids = _return_edge_labels_2d(input_map[i])
            full_union.extend(_ids)
    else:
        full_union = _return_edge_labels_2d(input_map)

    return list(full_union)


def remove_edge_labels(input_map: NDArray) -> NDArray:
    """Remove labels touching edges of array.

    Args:
        input_map: Input segmentation array (2D or 3D)

    Returns:
        Array with edge-touching labels removed
    """
    edge_labels = _return_edge_labels(input_map)
    cleaned_map = np.where(np.isin(input_map, edge_labels), 0, input_map)
    return cleaned_map


def shift_labels(
    input_map: NDArray,
    shift: int,
    return_shifted_labels: bool = False,
    remove_edge_labels: bool = True,
    dtype=np.uint64,
) -> tuple[NDArray, list[int]]:
    """Shift label values by specified amount.

    Args:
        input_map: Input segmentation array
        shift: Value to increment labels by
        return_shifted_labels: Return edge labels after shifting
        remove_edge_labels: Remove edge-touching labels

    Returns:
        Tuple containing:
            - Array with shifted labels
            - List of edge-touching labels

    Example:
        >>> input_map = np.array([[1, 0, 0], [0, 2, 0], [0, 0, 3]])
        >>> shifted_map, edge_labels = shift_labels(input_map, 10)
        >>> print(shifted_map)
        array([[11,  0,  0],
            [ 0, 12,  0],
            [ 0,  0, 13]])
        >>> print(edge_labels)
        [11, 13]
    """
    imap = input_map.copy()
    shifted_map = np.where(imap == 0, 0, imap + shift)

    edge_label: list[int] = []

    if len(imap.shape) == 2:
        edge_label += _return_edge_labels(imap)
    else:
        for dimension in imap:
            edge_label += _return_edge_labels(dimension)

    if return_shifted_labels:
        edge_label = [label + shift for label in edge_label]

    if remove_edge_labels:
        if return_shifted_labels:
            shifted_map = np.where(np.isin(shifted_map, edge_label), 0, shifted_map)
        else:
            _edge_label = [label + shift for label in edge_label]
            shifted_map = np.where(np.isin(shifted_map, _edge_label), 0, shifted_map)

    return shifted_map.astype(dtype), list(set(edge_label))


@njit(parallel=True)
def _remove_classes(label_in: NDArray, to_remove: list[int], background: int = 0, reindex: bool = False) -> NDArray:
    """Remove specified classes from labeled array.

    Args:
        label_in: Input labeled array
        to_remove: List of labels to remove
        background: Background label value
        reindex: Reindex remaining labels consecutively

    Returns:
        Array with specified classes removed/reindexed

    Example:
        >>> label_in = np.array([[1, 2, 1], [1, 0, 2], [0, 2, 3]])
        >>> _remove_classes(label_in, [1, 3])
        array([[0, 2, 0],
               [0, 0, 2],
               [0, 2, 0]])
    """
    label = label_in.copy()
    remove_set = set(to_remove)
    library = np.zeros(np.max(label) + 1, dtype="int")

    carry = 0
    for current_class in range(len(library)):
        if current_class in remove_set:
            library[current_class] = background
            if reindex:
                carry -= 1
        else:
            library[current_class] = current_class + carry

    for y in prange(len(label)):
        for x in prange(len(label[0])):
            current_label = label[y, x]
            if current_label != background:
                label[y, x] = library[current_label]

    return label


def remove_classes(label_in: NDArray, to_remove: list[int], background: int = 0, reindex: bool = False) -> NDArray:
    """Remove specified classes from labeled array.

    Args:
        label_in: Input labeled array
        to_remove: List of labels to remove
        background: Background label value
        reindex: Reindex remaining labels consecutively

    Returns:
        Array with specified classes removed/reindexed
    """
    return _remove_classes(label_in, to_remove, background=background, reindex=reindex)


@njit(parallel=True)
def contact_filter_lambda(label: NDArray, background: int = 0) -> NDArray:
    """Calculate contact proportion of classes with background.

    Args:
        label: Input labeled array
        background: Background label value

    Returns:
        Array of background contact proportions for each class

    Example:
        >>> label = np.array([[0, 1, 1], [0, 2, 1], [0, 0, 2]])
        >>> contact_filter_lambda(label)
        array([1.0, 1.0, 0.6])
    """
    num_labels = np.max(label)
    background_matrix = np.ones((num_labels + 1, 2), dtype="int")

    for y in range(1, len(label) - 1):
        for x in range(1, len(label[0]) - 1):
            current_label = label[y, x]

            background_matrix[current_label, 0] += int(label[y - 1, x] == background)
            background_matrix[current_label, 0] += int(label[y, x - 1] == background)
            background_matrix[current_label, 0] += int(label[y + 1, x] == background)
            background_matrix[current_label, 0] += int(label[y, x + 1] == background)

            background_matrix[current_label, 1] += int(label[y - 1, x] != current_label)
            background_matrix[current_label, 1] += int(label[y, x - 1] != current_label)
            background_matrix[current_label, 1] += int(label[y + 1, x] != current_label)
            background_matrix[current_label, 1] += int(label[y, x + 1] != current_label)

    proportion = background_matrix[:, 0] / background_matrix[:, 1]
    return proportion


def contact_filter(inarr: NDArray, threshold: float = 1.0, reindex: bool = False, background: int = 0) -> NDArray:
    """Filter array based on label contact with background.

    Args:
        inarr: Input labeled array
        threshold: Minimum background contact proportion for retention
        reindex: Reindex remaining labels consecutively
        background: Background label value

    Returns:
        Filtered labeled array

    Example:
        >>> inarr = np.array([[0, 1, 1], [0, 2, 1], [0, 0, 2]])
        >>> contact_filter(inarr, threshold=0.5)
        array([[0, 1, 1],
               [0, 0, 1],
               [0, 0, 0]])
    """
    label = inarr.copy()
    background_contact = contact_filter_lambda(label, background=background)
    to_remove = np.argwhere(background_contact < threshold).flatten()
    to_remove = np.delete(to_remove, np.where(to_remove == background))

    if len(to_remove) > 0:
        label = remove_classes(label, nb.typed.List(to_remove), reindex=reindex)

    return label


@njit
def _class_size(mask: NDArray, debug: bool = False, background: int = 0) -> tuple[NDArray, NDArray]:
    """Compute size of each class in mask.

    Args:
        mask: Input mask containing classes
        debug: Debug flag
        background: Background label value

    Returns:
        Tuple containing:
            - Array of class centroids
            - Array of class sizes

    Example:
        >>> mask = np.array([[0, 1, 1], [0, 2, 1], [0, 0, 2]])
        >>> _class_size(mask)
        (array([[       nan,        nan],
                [0.33333333, 1.66666667],
                [1.5       , 1.5       ]]),
        array([nan,  3.,  2.]))
    """
    cell_ids = list(np.unique(mask).flatten())
    if 0 in cell_ids:
        cell_ids.remove(background)
    cell_ids = np.array(cell_ids)

    min_cell_id = np.min(cell_ids)

    if min_cell_id != 1:
        mask = _numba_subtract(mask, min_cell_id - 1)

    num_classes = np.max(mask) + 1
    mean_sum = np.zeros((num_classes, 2))
    length = np.zeros((num_classes, 1))
    rows, cols = mask.shape

    for row in range(rows):
        for col in range(cols):
            return_id = mask[row, col]
            if return_id != background:
                mean_sum[return_id] += np.array([row, col], dtype=np.uint32)
                length[return_id][0] += 1

    mean_arr = np.divide(mean_sum, length)
    length[background][0] = np.nan
    mean_arr[background] = np.nan

    return mean_arr, length.flatten()


def size_filter(label: NDArray, limits: list[int] | None = None, reindex: bool = False) -> NDArray:
    """Filter classes based on size.

    Args:
        label: Input labeled array
        limits: [min_size, max_size] in pixels
        reindex: Reindex remaining labels consecutively

    Returns:
        Filtered labeled array

    Example:
        >>> label = np.array([[0, 1, 1], [0, 2, 1], [0, 0, 2]])
        >>> size_filter(label, limits=[1, 2])
        array([[0, 0, 0],
               [0, 2, 0],
               [0, 0, 2]])
    """
    if limits is None:
        limits = [0, 100000]

    _, points_class = _class_size(label)

    below = np.argwhere(points_class < limits[0]).flatten()
    above = np.argwhere(points_class > limits[1]).flatten()
    to_remove = list(below) + list(above)

    if len(to_remove) > 0:
        label = remove_classes(label, nb.typed.List(to_remove), reindex=reindex)

    return label


@njit
def numba_mask_centroid(
    mask: NDArray, skip_background: bool = True
) -> tuple[NDArray | None, NDArray | None, NDArray | None]:
    """Calculate centroids of labeled regions.

    Args:
        mask: Input mask containing classes
        skip_background: Skip background class calculation

    Returns:
        Tuple containing:
            - Array of (y,x) centroid coordinates
            - Array of region sizes
            - Array of region IDs
        Returns (None, None, None) if no cells found

    Example:
        >>> mask = np.array([[0, 1, 1], [0, 2, 1], [0, 0, 2]])
        >>> centers, sizes, ids = numba_mask_centroid(mask)
    """
    cell_ids = list(np.unique(mask).flatten())
    if 0 in cell_ids:
        cell_ids.remove(0)
    cell_ids = np.array(cell_ids)

    min_cell_id = np.min(cell_ids)

    if min_cell_id == 0:
        print("no cells in image. Only contains background with value 0.")
        return None, None, None

    num_classes = int(np.max(mask) if skip_background else np.max(mask) + 1)

    points_class = np.zeros((num_classes,), dtype=nb.uint32)
    center = np.zeros((num_classes, 2))
    ids = np.full((num_classes,), np.nan)

    if skip_background:
        for y in range(len(mask)):
            for x in range(len(mask[0])):
                class_id = mask[y, x]
                if class_id != 0:
                    points_class[class_id - 1] += 1
                    center[class_id - 1] += np.array([x, y])
                    ids[class_id - 1] = class_id
    else:
        for y in range(len(mask)):
            for x in range(len(mask[0])):
                class_id = mask[y, x]
                points_class[class_id] += 1
                center[class_id] += np.array([x, y])
                ids[class_id] = class_id

    x = center[:, 0] / points_class
    y = center[:, 1] / points_class
    center = np.stack((y, x)).T

    # Remove background and invalid IDs
    valid_mask = ~np.isnan(ids)
    center = center[valid_mask]
    points_class = points_class[valid_mask]
    ids = ids[valid_mask]

    bg_mask = ids != 0
    center = center[bg_mask]
    points_class = points_class[bg_mask]
    ids = ids[bg_mask]

    return center, points_class, ids.astype(np.uint32)


@nb.jit(nopython=True)
def sc_any(array: NDArray) -> bool:
    """Short-circuiting replacement for np.any().

    Args:
        array: Input array to check

    Returns:
        True if any element is True, False otherwise
    """
    for x in array.flat:
        if x:
            return True
    return False


@nb.jit(nopython=True)
def sc_all(array: NDArray) -> bool:
    """Short-circuiting replacement for np.all().

    Args:
        array: Input array to check

    Returns:
        True if all elements are True, False otherwise
    """
    for x in array.flat:
        if not x:
            return False
    return True


def remap_mask(input_mask: np.ndarray) -> np.ndarray:
    """Remap mask to have consecutive labels starting from 1.

    Args:
        input_mask: Input mask with non-consecutive labels

    Returns:
        Remapped mask with consecutive labels

    Example:
        >>> input_mask = np.array([[5, 6, 6], [5, 3, 6], [5, 5, 3]])
        >>> remap_mask(input_mask)
        array([[2, 3, 3],
               [2, 1, 3],
               [2, 2, 1]])
    """
    # Create lookup table as an array
    max_label = np.max(input_mask)
    lookup_array = np.zeros(max_label + 1, dtype=np.int32)

    cell_ids = np.unique(input_mask)[1:]
    lookup_table = dict(zip(cell_ids, range(1, len(cell_ids) + 1), strict=True))

    # Populate lookup array based on the dictionary
    for old_id, new_id in lookup_table.items():
        lookup_array[old_id] = new_id

    # Apply the mapping using NumPy indexing
    remapped_mask = lookup_array[input_mask]

    return remapped_mask



### Definition of default values used throughout the scPortrait pipeline ###
from typing import TypeAlias

import numpy as np

ChunkSize2D: TypeAlias = tuple[int, int]
ChunkSize3D: TypeAlias = tuple[int, int, int]

DEFAULT_LOG_NAME: str = "processing.log"
DEFAULT_FORMAT: str = "%d/%m/%Y %H:%M:%S"

DEFAULT_CONFIG_NAME: str = "config.yml"
DEFAULT_INPUT_IMAGE_NAME: str = "input_image"
DEFAULT_SDATA_FILE: str = "scportrait.sdata"

DEFAULT_PREFIX_MAIN_SEG: str = "seg_all"
DEFAULT_PREFIX_FILTERED_SEG: str = "seg_filtered"
DEFAULT_PREFIX_SELECTED_SEG: str = "seg_selected"

DEFAULT_SEG_NAME_0: str = "nucleus"
DEFAULT_SEG_NAME_1: str = "cytosol"

DEFAULT_CENTERS_NAME: str = "centers"
DEFAULT_CHUNK_SIZE_3D: ChunkSize3D = (1, 2000, 2000)
DEFAULT_CHUNK_SIZE_2D: ChunkSize2D = (2000, 2000)
DEFAULT_SCALE_FACTORS: list[int] = [2, 4, 8]

DEFAULT_SEGMENTATION_DIR_NAME: str = "segmentation"
DEFAULT_TILES_FOLDER: str = "tiles"
DEFAULT_FILTER_ADDTIONAL_FILE: str = "needs_additional_filtering.txt"

DEFAULT_EXTRACTION_DIR_NAME: str = "extraction"
DEFAULT_DATA_DIR: str = "data"
DEFAULT_NAME_SINGLE_CELL_IMAGES = "single_cell_images"
IMAGE_DATACONTAINER_NAME = f"obsm/{DEFAULT_NAME_SINGLE_CELL_IMAGES}"
DEFAULT_CELL_ID_NAME = "scportrait_cell_id"
INDEX_DATACONTAINER_NAME = f"obs/{DEFAULT_CELL_ID_NAME}"

DEFAULT_IMAGE_DTYPE: np.dtype = np.uint16
DEFAULT_SEGMENTATION_DTYPE: np.dtype = np.uint64
DEFAULT_SINGLE_CELL_IMAGE_DTYPE: np.dtype = np.float16

DEFAULT_SEGMENTATION_FILE: str = "segmentation.h5"
DEFAULT_CLASSES_FILE: str = "classes.csv"
DEFAULT_REMOVED_CLASSES_FILE: str = "removed_classes.csv"
DEFAULT_EXTRACTION_FILE: str = "single_cells.h5sc"
DEFAULT_BENCHMARKING_FILE: str = "benchmarking.csv"

DEFAULT_SELECTION_DIR_NAME: str = "selection"
DEFAULT_FEATURIZATION_DIR_NAME: str = "featurization"

DEFAULT_CHANNELS_NAME: str = "channels"
DEFAULT_MASK_NAME: str = "labels"



"""
h5sc
=======

Functions to work with scPortrait's standardized single-cell data format.
"""

import numpy as np

from scportrait.pipeline._utils.constants import DEFAULT_CELL_ID_NAME, DEFAULT_NAME_SINGLE_CELL_IMAGES


def get_image_with_cellid(adata, cell_id: list[int] | int, select_channel: int | list[int] | None = None) -> np.ndarray:
    """Get single cell images from the cells with the provided cell IDs. Images are returned in the order of the cell IDs.

    Args:
        adata: An AnnData object with obsm["single_cell_images"] containing a memory-backed array of the single-cell images.
        cell_id: The cell ID of the cell to retrieve the image for.
        select_channel: The channel to select from the image. If `None`, all channels are returned.

    Returns:
        The image(s) of the cell with the passed Cell IDs.
    """
    lookup = dict(zip(adata.obs[DEFAULT_CELL_ID_NAME], adata.obs.index.astype(int), strict=True))
    image_container = adata.obsm[DEFAULT_NAME_SINGLE_CELL_IMAGES]

    if isinstance(cell_id, int):
        cell_id = [cell_id]

    for x in cell_id:
        assert x in lookup.keys(), f"CellID {x} is not present in the AnnData object."

    images = []
    for _id in cell_id:
        idx = lookup[_id]
        if select_channel is None:
            image = image_container[idx][:]
        else:
            image = image_container[idx][select_channel]
        images.append(image)

    array = np.array(images)
    if array.shape[0] == 1:  # Check if the first dimension is 1
        return array.squeeze(axis=0)  # Remove the first dimension
    else:
        return array



from __future__ import annotations

import os
import shutil
import tempfile
from pathlib import Path
from random import choice
from string import ascii_lowercase
from typing import Literal

import requests
from filelock import FileLock
from tqdm import tqdm


def _download(
    url: str,
    archive_format: Literal["zip", "tar", "tar.gz", "tgz"] = None,
    output_file_name: str = None,
    output_path: str | Path = None,
    block_size: int = 1024,
    overwrite: bool = False,
) -> None:  # pragma: no cover
    """Downloads a file irrespective of format.

    Args:
        url: URL to download.
        archive_format: The format if an archive file.
        output_file_name: Name of the downloaded file.
        output_path: Path to download/extract the files to. Defaults to 'OS tmpdir' if not specified.
        block_size: Block size for downloads in bytes.
        overwrite: Whether to overwrite existing files.
    """
    if output_file_name is None:
        letters = ascii_lowercase
        output_file_name = f"scportrait_tmp_{''.join(choice(letters) for _ in range(10))}"

    if output_path is None:
        output_path = tempfile.gettempdir()

    def _sanitize_file_name(file_name):
        if os.name == "nt":
            file_name = file_name.replace("?", "_").replace("*", "_")
        return file_name

    download_to_path = Path(
        _sanitize_file_name(
            f"{output_path}{output_file_name}"
            if str(output_path).endswith("/")
            else f"{output_path}/{output_file_name}"
        )
    )

    Path(output_path).mkdir(parents=True, exist_ok=True)
    lock_path = f"{download_to_path}.lock"

    with FileLock(lock_path):
        if download_to_path.exists():
            warning = f"File {download_to_path} already exists!"
            if not overwrite:
                print(warning)
                Path(lock_path).unlink()
                return
            else:
                print(f"{warning} Overwriting...")

        response = requests.get(url, stream=True)
        total = int(response.headers.get("content-length", 0))

        temp_file_name = f"{download_to_path}.part"

        with (
            open(temp_file_name, "wb") as file,
            tqdm(total=total, unit="B", unit_scale=True, desc="Downloading...") as progress_bar,
        ):
            for data in response.iter_content(block_size):
                file.write(data)
                progress_bar.update(len(data))

        Path(temp_file_name).replace(download_to_path)

        if archive_format:
            shutil.unpack_archive(download_to_path, output_path, format=archive_format)
            os.remove(download_to_path)
        else:
            download_to_path.with_name(output_file_name)

    Path(lock_path).unlink()



"""Collection of plotting functions for scPortrait's standardized single-cell image format"""

import warnings
from collections.abc import Iterable

import matplotlib.pyplot as plt
import numpy as np
from anndata import AnnData
from matplotlib.axes import Axes
from matplotlib.figure import Figure
from mpl_toolkits.axes_grid1.inset_locator import inset_axes

from scportrait.pipeline._utils.constants import DEFAULT_CELL_ID_NAME
from scportrait.tools.h5sc import get_image_with_cellid


def _reshape_image_array(arr: np.ndarray) -> np.ndarray:
    """
    Reshape an array from (n, c, x, y) to (n*c, x, y) while maintaining the order:
    n1 c1, n1 c2, ..., n1 cy, n2 c1, ..., nx cy.
    """
    if len(arr.shape) == 3:
        return arr
    else:
        n, c, x, y = arr.shape
        return arr.reshape(n * c, x, y)


def _plot_image_grid(
    ax: Axes,
    images: np.ndarray,
    nrows: int,
    ncols: int,
    spacing: float = 0.01,
    image_titles: list[str] | None = None,
    image_titles_fontsize: int = 10,
    row_labels: list[str] | None = None,
    col_labels: list[str] | None = None,
    axs_title: str | None = None,
    axs_title_padding: float = 0,
    axs_title_fontsize: int = 12,
    cmap="viridis",
    vmin: float = 0,
    vmax: float = 1,
) -> None:
    """Helper function to plot an image grid with consistent spacing between rows and columns.

    Args:
        ax: The matplotlib axes object to plot on.
        images: The images to plot.
        nrows: The number of rows in the grid.
        ncols: The number of columns in the grid.
        spacing: The spacing between cells in the grid expressed as fraction of the cell image size.
        image_titles: The titles for each image in the grid.
        image_titles_fontsize: The fontsize of the image titles.
        row_labels: The labels for each row in the grid.
        col_labels: The labels for each column in the grid.
        axs_title: The title of the axes.
        axs_title_padding: The padding of the axes title.
        axs_title_fontsize: The fontsize of the axes title.
        cmap: The colormap to use for the images.
        vmin: The minimum value for the colormap.
        vmax: The maximum value for the colormap.

    Returns:
        None
    """

    ax.set_title(axs_title, fontsize=axs_title_fontsize, pad=axs_title_padding)
    ax.axis("off")

    # Adjust row spacing if image titles are provided
    title_adjustment = 0.01 if image_titles is not None else 0

    # Calculate effective cell width and height considering spacing
    spacing = spacing / max(nrows, ncols)
    cell_width = (1 - (ncols + 1) * spacing) / ncols
    cell_height = (1 - (nrows + 1) * spacing - title_adjustment) / nrows  # Adjust height

    if len(images) > nrows * ncols:
        warnings.warn(
            f"Number of images exceeds grid size. Only the first {nrows * ncols} images will be displayed.",
            stacklevel=2,
        )
        images = images[: nrows * ncols]

    for i, img in enumerate(images):
        row = i // ncols
        col = i % ncols

        ax_sub = ax.inset_axes(
            [
                spacing + col * (cell_width + spacing),  # X position
                1 - (spacing + (row + 1) * (cell_height + spacing + title_adjustment)),  # Y position
                cell_width,  # Width
                cell_height,  # Height
            ]
        )

        ax_sub.imshow(img, cmap=cmap, vmin=vmin, vmax=vmax)
        ax_sub.yaxis.set_visible(False)
        ax_sub.xaxis.set_visible(False)

        if row_labels is not None and col == 0:
            ax_sub.set_ylabel(row_labels[row], fontsize=image_titles_fontsize)
            ax_sub.yaxis.set_visible(True)
            ax_sub.tick_params(left=False, labelleft=False)

        if col_labels is not None and row == 0:
            ax_sub.set_title(f"{col_labels[col]}", fontsize=image_titles_fontsize)

        if image_titles is not None:
            ax_sub.set_title(f"{image_titles[i]}", fontsize=image_titles_fontsize, pad=2)


def cell_grid_single_channel(
    adata,
    select_channel: int | str,
    n_cells: int = 16,
    cell_ids: int | list[int] | None = None,
    cell_labels: list[str] | None = None,
    show_cell_id: bool = False,
    title: str | None = None,
    show_title: bool = True,
    cmap="viridis",
    ncols: int | None = None,
    nrows: int | None = None,
    single_cell_size: int = 2,
    spacing: float = 0.025,
    ax: Axes = None,
    return_fig: bool = False,
    show_fig: bool = True,
) -> None | Figure:
    """Visualize a single channel from h5sc object in a grid.

    Args:
        adata: An scPortrait single-cell image dataset.
        select_channel: The channel to visualize.
        n_cells: The number of cells to visualize. This number of cells will randomly be selected. If `None`, `cell_ids` must be provided.
        cell_ids: cell IDs for the specific cells that should be visualized. If `None`, `n_cells` must be provided.
        cell_labels: Label to plot as title for each single-cell image if provided.
        show_cell_id: Whether to show the cell ID as title for each single-cell image. Can not be used together with `cell_labels`.
        title: The title of the plot.
        show_title: Whether to show the title.
        cmap: The colormap to use for the images.
        ncols: The number of columns in the grid. If not specified will be automatically calculated to make a square grid.
        nrows: The number of rows in the grid. If not specified will be automatically calculated to make a square grid.
        single_cell_size: The size of each cell in the grid.
        spacing: The spacing between cells in the grid expressed as fraction of the cell image size.
        ax: The matplotlib axes object to plot on. If `None`, a new figure is created.
        return_fig: If `True`, the function returns the figure object instead of displaying it.
        show_fig: If `True`, the figure is displayed.

    Returns:
        If `return_fig=True`, the figure object is returned. Otherwise, the figure is displayed.
    """
    # ensure that cell_ids are passed as a list
    if isinstance(cell_ids, int):
        _cell_ids = [cell_ids]
    else:
        _cell_ids = cell_ids

    # ensure that this parameter does not need to be set
    if _cell_ids is not None:
        n_cells = None

    if isinstance(select_channel, str):
        channel_id = adata.uns["single_cell_images"]["channel_names"].tolist().index(select_channel)
        channel_name = select_channel
    elif isinstance(select_channel, int):
        channel_id = select_channel
        channel_name = adata.uns["single_cell_images"]["channel_names"].tolist()[select_channel]

    if n_cells is None:
        assert _cell_ids is not None, "Either `n_cells` or `cell_ids` must be provided."
        n_cells = len(_cell_ids)
    else:
        # get random cells if no specific IDs are provided
        _cell_ids = adata.obs[DEFAULT_CELL_ID_NAME].sample(n_cells).values
        assert cell_labels is None, "can not provide labels for randomly sampled cells."

    if cell_labels is not None:
        assert show_cell_id is False, "can not show cell IDs and labels at the same time."
        spacing = 0.05

    if show_cell_id:
        cell_labels = [f"Cell ID {x}" for x in _cell_ids]
        spacing = 0.05

    # configure size for resulting single-cell image grid
    if ncols is None and nrows is None:
        ncols = int(np.ceil(np.sqrt(n_cells)))
        nrows = int(np.ceil(n_cells / ncols))
    elif ncols is None:
        ncols = int(np.ceil(n_cells / nrows))
    elif nrows is None:
        nrows = int(np.ceil(n_cells / ncols))

    # set up image size
    fig_height = nrows * single_cell_size
    fig_width = ncols * single_cell_size

    # configure title
    if title is None:
        title = f"Channel {channel_name}"
    if not show_title:
        title = None

    # create figure object
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(fig_width, fig_height))
    else:
        fig = ax.get_figure()

    spacing = spacing * single_cell_size
    images = get_image_with_cellid(adata, _cell_ids, channel_id)
    _plot_image_grid(
        ax, images, nrows=nrows, ncols=ncols, axs_title=title, image_titles=cell_labels, cmap=cmap, spacing=spacing
    )

    if return_fig:
        return fig
    elif show_fig:
        plt.show()
        return None
    else:
        return None


def cell_grid_multi_channel(
    adata,
    n_cells: int = 5,
    cell_ids: int | list[int] | None = None,
    select_channels: list[int] | list[str] | None = None,
    title: str | None = None,
    show_cell_id: bool = True,
    label_channels: bool = True,
    row_labels: list[str] | None = None,
    cmap="viridis",
    spacing: float = 0.025,
    single_cell_size: int = 2,
    ax: Axes = None,
    return_fig: bool = False,
    show_fig: bool = True,
) -> None | Figure:
    """Plot a grid of single-cell images where each row is a unique cell and each column contains a different channel.

    Args:
        adata: An scPortrait single-cell image dataset.
        n_cells: The number of cells to visualize. This number of cells will randomly be selected. If `None`, `cell_ids` must be provided.
        cell_ids: cell IDs for the specific cells that should be visualized. If `None`, `n_cells` must be provided.
        select_channels: The channels to visualize. Channels are identified as int values corresponding to their order in the h5sc dataset. If `None`, all channels will be visualized.
        title: The title of the plot.
        show_cell_id: Whether to show the cell ID as row label for each cell in the image grid.
        label_channels: Whether to show the channel names as titles for column in the image grid.
        row_labels: can override the labels plotted on the y-axis on the first element of each row. If using not compatible with passing `show_cell_id` as `True`.
        cmap: The colormap to use for the images.
        spacing: The spacing between cells in the grid expressed as fraction of the cell image size.
        single_cell_size: The size of each cell in the grid.
        ax: The matplotlib axes object to plot on. If `None`, a new figure is created.
        return_fig: If `True`, the function returns the figure object instead of displaying it.
        show_fig: If `True`, the figure is displayed.

    Returns:
        If `return_fig=True`, the figure object is returned. Otherwise, the figure is displayed.
    """
    # ensure that cell_ids are passed as a list
    if isinstance(cell_ids, Iterable):
        _cell_ids = cell_ids
    else:
        _cell_ids = [cell_ids]

    if cell_ids is not None:
        n_cells = None

    if n_cells is None:
        assert _cell_ids is not None, "Either `n_cells` or `cell_ids` must be provided."
        n_cells = len(_cell_ids)
    else:
        # Get random cells if no specific IDs are provided
        _cell_ids = adata.obs[DEFAULT_CELL_ID_NAME].sample(n_cells).values

    if row_labels is not None:
        assert (
            show_cell_id is False
        ), "If manually providing row_labels, can not automatically annotate rows with cell IDs. Set `show_cell_id` to False."
        assert len(row_labels) == n_cells, "Length of `row_labels` must match the number of cells to be visualized."
    else:
        row_labels = [f"cell ID {_id}" for _id in _cell_ids] if show_cell_id else None

    n_channels = adata.uns["single_cell_images"]["n_channels"]
    channel_names = adata.uns["single_cell_images"]["channel_names"]

    # Collect images in a list
    images = get_image_with_cellid(adata, _cell_ids)
    if select_channels is not None:
        if not isinstance(select_channels, Iterable):
            select_channels = [select_channels]
        if np.all(isinstance(x, str) for x in select_channels):
            select_channels = [list(channel_names).index(x) for x in select_channels]
        images = images[:, select_channels, :, :]
        channel_names = [channel_names[i] for i in select_channels]
        n_channels = len(select_channels)
    images = _reshape_image_array(images)

    # Determine grid size
    nrows = n_cells
    ncols = n_channels

    # Configure plot size
    fig_width = ncols * single_cell_size
    fig_height = nrows * single_cell_size

    # Create figure object
    if ax is None:
        fig, ax = plt.subplots(1, 1, figsize=(fig_width, fig_height))
    else:
        fig = ax.get_figure()

    # Call the image grid function
    spacing = spacing * single_cell_size
    _plot_image_grid(
        ax=ax,
        images=images,
        nrows=nrows,
        ncols=ncols,
        spacing=spacing,
        row_labels=row_labels,
        col_labels=channel_names if label_channels else None,
        axs_title=title,
        cmap=cmap,
        vmin=0,
        vmax=1,
    )

    if return_fig:
        return fig
    elif show_fig:
        plt.show()
        return None
    else:
        return None


def cell_grid(
    adata: AnnData,
    select_channel: int | str | list[int] | list[str] = None,
    n_cells: int | None = None,
    cell_ids: int | list[int] | None = None,
    show_cell_id: bool = True,
    cmap="viridis",
    return_fig: bool = False,
    show_fig: bool = True,
):
    """Visualize single-cell images of cells in an AnnData object.

    Uses either `cell_grid_single_channel` or `cell_grid_multi_channel` depending on the input.
    Use these functions if you would like more control over the visualization.

    Args:
        adata: An scPortrait single-cell image dataset.
        select_channel: The channel to visualize. Can be either a single int value or a list of values. If None all channels will  be visualized.
        n_cells: The number of cells to visualize. This number of cells will randomly be selected. If set to `None`, `cell_ids` must be provided.
        cell_ids: cell IDs of the cells that are to be visualiazed. If `None`, `n_cells` must be provided.
        show_cell_ids: Whether to show the cell ID as titles/y-labels for each single-cell image.
        cmap: The colormap to use for the images.
        return_fig: If `True`, the function returns the figure object instead of displaying it.
        show_fig: If `True`, the figure is displayed.

    Returns:
        If `return_fig=True`, the figure object is returned. Otherwise, the figure is displayed.
    """
    if isinstance(select_channel, int | str):
        if cell_ids is None:
            if n_cells is None:
                n_cells = 16
        return cell_grid_single_channel(
            adata,
            select_channel,
            n_cells=n_cells,
            cell_ids=cell_ids,
            show_cell_id=show_cell_id,
            cmap=cmap,
            return_fig=return_fig,
            show_fig=show_fig,
        )
    else:
        if cell_ids is None:
            if n_cells is None:
                n_cells = 5
        return cell_grid_multi_channel(
            adata,
            n_cells=n_cells,
            cell_ids=cell_ids,
            cmap=cmap,
            select_channels=select_channel,
            show_cell_id=show_cell_id,
            return_fig=return_fig,
            show_fig=show_fig,
        )



import csv
import gc
import os
import shutil
import traceback
from collections import defaultdict
from multiprocessing import Pool
from pathlib import Path
from typing import Any

import h5py
import numpy as np
import numpy.typing as npt
from alphabase.io import tempmmap
from tqdm.auto import tqdm

from scportrait.pipeline._base import ProcessingStep
from scportrait.pipeline._utils.segmentation import sc_any


class SegmentationFilter(ProcessingStep):
    """SegmentationFilter helper class used for creating workflows to filter generated segmentation masks before extraction."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        self.identifier: int | None = None
        self.window: tuple[slice, slice] | None = None
        self.input_path: str | Path | None = None

    def read_input_masks(self, input_path: str | Path) -> npt.NDArray[np.uint16]:
        """Read input masks from a given HDF5 file.

        Args:
            input_path: Path to the HDF5 file containing the input masks.

        Returns:
            Array containing the input masks.
        """
        with h5py.File(input_path, "r") as hf:
            hdf_input = hf.get("labels")
            input_masks = tempmmap.array(
                shape=hdf_input.shape,
                dtype=np.uint16,
                tmp_dir_abs_path=self._tmp_dir_path,
            )
            input_masks = hdf_input[:2, :, :]
        return input_masks

    def save_classes(self, classes: dict[str, list[Any]]) -> None:
        """Save the filtered classes to a CSV file.

        Args:
            classes: Dictionary of classes to save.
        """
        filtered_path = Path(self.directory) / self.DEFAULT_REMOVED_CLASSES_FILE
        to_write = "\n".join([f"{str(x)}:{str(y)}" for x, y in classes.items()])
        with open(filtered_path, "w") as myfile:
            myfile.write(to_write)
        self.log(f"Saved nucleus_id:cytosol_id matchings of all cells that passed filtering to {filtered_path}.")

    def initialize_as_tile(
        self, identifier: int, window: tuple[slice, slice], input_path: str | Path, zarr_status: bool = True
    ) -> None:
        """Initialize Filtering Step with further parameters needed for filtering segmentation results.

        Important:
            This function is intended for internal use by the :class:`TiledFilterSegmentation` helper class.
            In most cases it is not relevant to the creation of custom filtering workflows.

        Args:
            identifier: Unique index of the tile.
            window: Defines the window which is assigned to the tile. The window will be applied
                   to the input. The first element refers to the first dimension of the image and so on.
            input_path: Location of the input HDF5 file. During tiled segmentation the
                       :class:`TiledSegmentation` derived helper class will save the input image
                       in form of a HDF5 file.
            zarr_status: Status of zarr saving, by default True.
        """
        self.identifier = identifier
        self.window = window
        self.input_path = input_path
        self.save_zarr = zarr_status

    def call_as_tile(self) -> None:
        """Wrapper function for calling segmentation filtering on an image tile.

        Important:
            This function is intended for internal use by the :class:`TiledSegmentation` helper class.
            In most cases it is not relevant to the creation of custom segmentation workflows.
        """
        with h5py.File(self.input_path, "r") as hf:
            hdf_input = hf.get("labels")
            if hdf_input is None:
                raise ValueError("No 'labels' dataset found in HDF5 file")

            c, _, _ = hdf_input.shape
            x1 = self.window[0].start
            x2 = self.window[0].stop
            y1 = self.window[1].start
            y2 = self.window[1].stop

            if any(v is None for v in [x1, x2, y1, y2]):
                raise ValueError("Window slice boundaries cannot be None")

            x = x2 - x1
            y = y2 - y1

            input_image = tempmmap.array(shape=(2, x, y), dtype=np.uint16, tmp_dir_abs_path=self._tmp_dir_path)
            input_image = hdf_input[:2, self.window[0], self.window[1]]

        if sc_any(input_image):
            try:
                self.log(f"Beginning filtering on tile in position [{self.window[0]}, {self.window[1]}]")
                super().__call__(input_image)
            except (OSError, ValueError, RuntimeError) as e:
                self.log(f"An error occurred: {e}")
                self.log(traceback.format_exc())
        else:
            print(f"Tile in position [{self.window[0]}, {self.window[1]}] only contained zeroes.")
            try:
                super().__call_empty__(input_image)
            except (OSError, ValueError, RuntimeError) as e:
                self.log(f"An error occurred: {e}")
                self.log(traceback.format_exc())

        del input_image
        gc.collect()

        self.log(f"Writing out window location to file at {self.directory}/window.csv")
        with open(f"{self.directory}/window.csv", "w") as f:
            f.write(f"{self.window}\n")
        self.log(f"Filtering of tile with the slicing {self.window} finished.")

    def get_output_path(self) -> Path:
        return Path(self.directory) / self.DEFAULT_SEGMENTATION_FILE


class TiledSegmentationFilter(SegmentationFilter):
    """TiledSegmentationFilter helper class used for creating workflows to filter generated segmentation masks using a tiled approach."""

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)
        if not hasattr(self, "method"):
            raise AttributeError("No SegmentationFilter method defined, please set attribute ``method``")
        self.tile_directory: Path | None = None

    def initialize_tile_list(
        self, tileing_plan: list[tuple[slice, slice]], input_path: str | Path
    ) -> list[SegmentationFilter]:
        """Initialize the list of tiles for segmentation filtering.

        Args:
            tileing_plan: List of windows defining the tiling plan.
            input_path: Path to the input HDF5 file.

        Returns:
            List of initialized tiles.
        """
        _tile_list = []
        self.input_path = input_path

        for i, window in enumerate(tileing_plan):
            local_tile_directory = Path(self.tile_directory) / str(i)
            current_tile = self.method(  # type: ignore
                self.config,
                local_tile_directory,
                project_location=self.project_location,
                debug=self.debug,
                overwrite=self.overwrite,
                intermediate_output=self.intermediate_output,  # type: ignore
            )
            current_tile.initialize_as_tile(i, window, self.input_path, zarr_status=False)  # type: ignore
            _tile_list.append(current_tile)

        return _tile_list

    def initialize_tile_list_incomplete(
        self, tileing_plan: list[tuple[slice, slice]], incomplete_indexes: list[int], input_path: str | Path
    ) -> list[SegmentationFilter]:
        """Initialize the list of incomplete tiles for segmentation filtering.

        Args:
            tileing_plan: List of windows defining the tiling plan.
            incomplete_indexes: List of indexes for incomplete tiles.
            input_path: Path to the input HDF5 file.

        Returns:
            List of initialized incomplete tiles.
        """
        _tile_list = []
        self.input_path = input_path

        for i, window in zip(incomplete_indexes, tileing_plan, strict=False):
            local_tile_directory = Path(self.tile_directory) / str(i)
            current_tile = self.method(  # type: ignore
                self.config,
                local_tile_directory,
                project_location=self.project_location,
                debug=self.debug,
                overwrite=self.overwrite,
                intermediate_output=self.intermediate_output,  # type: ignore
            )
            current_tile.initialize_as_tile(i, window, self.input_path, zarr_status=False)  # type: ignore
            _tile_list.append(current_tile)

        return _tile_list

    def calculate_tileing_plan(self, mask_size: tuple[int, int]) -> list[tuple[slice, slice]]:
        """Calculate the tiling plan based on the mask size.

        Args:
            mask_size: Size of the mask.

        Returns:
            List of windows defining the tiling plan.
        """
        tileing_plan_path = f"{self.directory}/tileing_plan.csv"

        if Path(tileing_plan_path).is_file():
            self.log(f"tileing plan already found in directory {tileing_plan_path}.")
            if self.overwrite:
                self.log("Overwriting existing tileing plan.")
                Path(tileing_plan_path).unlink()
            else:
                self.log("Reading existing tileing plan from file.")
                with Path(tileing_plan_path).open() as f:
                    tileing_plan = [eval(line) for line in f.readlines()]
                return tileing_plan

        _tileing_plan: list[tuple[slice, slice]] = []
        side_size = np.floor(np.sqrt(int(self.config["tile_size"])))
        tiles_side = np.round(np.array(mask_size) / side_size).astype(int)
        tile_size = np.array(mask_size) // tiles_side

        self.log(f"input image {mask_size[0]} px by {mask_size[1]} px")
        self.log(f"target_tile_size: {self.config['tile_size']}")
        self.log("tileing plan:")
        self.log(f"{tiles_side[0]} rows by {tiles_side[1]} columns")
        self.log(f"{tile_size[0]} px by {tile_size[1]} px")

        for y in range(tiles_side[0]):
            for x in range(tiles_side[1]):
                last_row = y == tiles_side[0] - 1
                last_column = x == tiles_side[1] - 1

                lower_y = y * tile_size[0]
                lower_x = x * tile_size[1]

                upper_y = (y + 1) * tile_size[0]
                upper_x = (x + 1) * tile_size[1]

                if last_row:
                    upper_y = mask_size[0]

                if last_column:
                    upper_x = mask_size[1]

                _tileing_plan.append((slice(lower_y, upper_y, None), slice(lower_x, upper_x, None)))

        with open(tileing_plan_path, "w") as f:
            for item in _tileing_plan:
                f.write(f"{item}\n")

        return _tileing_plan

    def execute_tile_list(self, tile_list: list[SegmentationFilter], n_cpu: int | None = None) -> list[Path]:
        """Execute the filtering process for a list of tiles.

        Args:
            tile_list: List of tiles to process.
            n_cpu: Number of CPU cores to use, by default None.

        Returns:
            List of output file paths for the processed tiles.
        """

        def f(x: SegmentationFilter) -> Path:
            try:
                x.call_as_tile()
            except (OSError, ValueError, RuntimeError) as e:
                self.log(f"An error occurred: {e}")
                self.log(traceback.format_exc())
            return x.get_output_path()

        if n_cpu == 1:
            self.log(f"Running sequentially on {n_cpu} CPU")
            return list(map(f, tile_list))
        else:
            n_processes = n_cpu if n_cpu else os.cpu_count()
            self.log(f"Running in parallel on {n_processes} CPUs")
            with Pool(n_processes) as pool:
                return list(pool.imap(f, tile_list))

    def execute_tile(self, tile: SegmentationFilter) -> Path:
        """Execute the filtering process for a single tile.

        Args:
            tile: Tile to process.

        Returns:
            Output file path for the processed tile.
        """
        tile.call_as_tile()
        return tile.get_output_path()

    def initialize_tile_directory(self) -> None:
        """Initialize the directory for storing tile outputs."""
        self.tile_directory = Path(self.directory) / self.DEFAULT_TILES_FOLDER
        if self.tile_directory.exists():
            self.log(f"Directory {self.tile_directory} already exists.")
            if self.overwrite:
                self.log("Overwriting existing tiles folder.")
                shutil.rmtree(self.tile_directory)
                self.tile_directory.mkdir()
            else:
                os.makedirs(self.tile_directory)
        else:
            self.tile_directory.mkdir()

    def collect_results(self) -> npt.NDArray[np.uint16]:
        """Collect the results from the processed tiles.

        Returns:
            Array containing the combined results from all tiles.
        """
        self.log("Reading in all tile results")
        with h5py.File(self.input_path, "r") as hf:
            hdf_input = hf.get("labels")
            if hdf_input is None:
                raise ValueError("No 'labels' dataset found in HDF5 file")
            c, y, x = hdf_input.shape

        self.log(f"Output image will have shape {c, y, x}")
        output_image = np.zeros((c, y, x), dtype=np.uint16)
        classes: defaultdict[str, list[Any]] = defaultdict(list)

        with open(f"{self.directory}/window.csv") as f:
            _window_locations = [eval(line.strip()) for line in f.readlines()]

        self.log(f"Expecting {len(_window_locations)} tiles")

        for i, loc in tqdm(enumerate(_window_locations), total=len(_window_locations)):
            out_dir = os.path.join(self.tile_directory, str(i))
            with h5py.File(f"{out_dir}/segmentation.h5", "r") as hf:
                data = hf.get("labels")
                if data is None:
                    raise ValueError(f"No 'labels' dataset found in tile {i}")
                for cls, mappings in csv.reader(open(f"{out_dir}/filtered_classes.csv")):
                    classes[cls].append(mappings)
                output_image[:, loc[0], loc[1]] = data[:, :]

        self.save_classes(classes)

        return output_image



import os
import timeit

import matplotlib.pyplot as plt
import numpy as np
from skfmm import travel_time as skfmm_travel_time
from skimage.color import label2rgb
from skimage.morphology import binary_erosion, dilation, disk
from skimage.segmentation import watershed

from scportrait.pipeline._utils.segmentation import (
    contact_filter,
    global_otsu,
    numba_mask_centroid,
    remove_edge_labels,
    segment_global_threshold,
    segment_local_threshold,
)
from scportrait.pipeline.segmentation.segmentation import (
    ShardedSegmentation,
)
from scportrait.pipeline.segmentation.workflows._base_segmentation_workflow import _BaseSegmentation


class _ClassicalSegmentation(_BaseSegmentation):
    def __init__(self, *args, **kwargs):
        self.maps = None  # just a typehint hack. Remove this and fix it properly.
        super().__init__(*args, **kwargs)

    def _visualize_final_masks(self):
        if self.segment_nuclei:
            image = label2rgb(
                self.maps["nucleus_segmentation"],
                self.maps["input_image"][0],
                bg_label=0,
                alpha=0.6,
            )

            fig = plt.figure(frameon=False, figsize=(10, 10))
            plt.imshow(image)
            plt.scatter(self.nucleus_centers[:, 1], self.nucleus_centers[:, 0], s=2, color="red")
            plt.axis("off")
            fig.tight_layout()
            fig.savefig(os.path.join(self.directory, "finalized_nucleus_mask.png"))

        if self.segment_cytosol:
            image = label2rgb(
                self.maps["cytosol_segmentation"],
                self.maps["input_image"][1],
                bg_label=0,
                alpha=0.6,
            )

            fig = plt.figure(frameon=False, figsize=(10, 10))
            plt.imshow(image)
            plt.scatter(self.nucleus_centers[:, 1], self.nucleus_centers[:, 0], s=2, color="red")
            plt.axis("off")
            fig.tight_layout()
            fig.savefig(os.path.join(self.directory, "cytosol_mask.png"))

    def _get_processing_parameters(self):
        # check if normalization is required
        if "lower_quantile_normalization" in self.config.keys():
            self.normalization_input_image = True
        elif "upper_quantile_normalization" in self.config.keys():
            self.normalization_input_image = True
        else:
            self.normalization_input_image = False
            self.log("No normalization will be performed")

        if self.normalization_input_image:
            self.log("Normalization of the input image will be performed.")

            # get lower quantile normalization range
            self.lower_quantile_normalization_input_image = 0
            self.upper_quantile_normalization_input_image = 1

            if isinstance(self.config["lower_quantile_normalization"], float):
                self.lower_quantile_normalization_input_image = self.config["lower_quantile_normalization"]

            elif isinstance(self.config["lower_quantile_normalization"], list):
                if len(self.config["lower_quantile_normalization"]) != self.input_image.shape[0]:
                    raise ValueError(
                        "Number of specified normalization ranges for lower quantile normalization does not match the number of input channels."
                    )
                self.lower_quantile_normalization_input_image = dict(
                    zip(
                        range(self.input_image.shape[0]),
                        self.config["lower_quantile_normalization"],
                        strict=False,
                    )
                )
            else:
                param = self.config["lower_quantile_normalization"]
                raise ValueError(
                    f"lower_quantile_normalization must be either a float or a list of floats. Instead recieved {param}"
                )

            # get upper quantile normalization range
            if isinstance(self.config["upper_quantile_normalization"], float):
                self.upper_quantile_normalization_input_image = self.config["upper_quantile_normalization"]

            elif isinstance(self.config["upper_quantile_normalization"], list):
                if len(self.config["upper_quantile_normalization"]) != self.input_image.shape[0]:
                    raise ValueError(
                        "Number of specified normalization ranges for upper quantile normalization does not match the number of input channels."
                    )
                self.upper_quantile_normalization_input_image = dict(
                    zip(
                        range(self.input_image.shape[0]),
                        self.config["upper_quantile_normalization"],
                        strict=False,
                    )
                )
            else:
                param = self.config["upper_quantile_normalization"]
                raise ValueError(
                    f"upper_quantile_normalization must be either a float or a list of floats. Instead recieved {param}"
                )

            # check that the normalization ranges are of the same type otherwise this will result in issues
            assert type(self.lower_quantile_normalization_input_image) == type(  # noqa: E721
                self.upper_quantile_normalization_input_image
            )  # these need to be the same types! So we need to circumvent the ruff linting rules here

        # check if median filtering is required
        if "median_filter_size" in self.config.keys():
            self.median_filtering_input_image = True
            self.median_filter_size_input_image = self.config["median_filter_size"]
        else:
            self.median_filtering_input_image = False

        # setup nucleus segmentation
        if "nucleus_segmentation" in self.config.keys():
            self.segment_nuclei = True

            # check for additional normalization before segmentation
            # set default values to 0 and 1 so no normalization occurs
            self.normalization_nucleus_lower_quantile = 0
            self.normalization_cytosol_upper_quantile = 1
            self.normalization_nucleus_segmentation = False

            if "lower_quantile_normalization" in self.config["nucleus_segmentation"]:
                self.normalization_nucleus_segmentation = True
                self.normalization_nucleus_lower_quantile = self.config["nucleus_segmentation"][
                    "lower_quantile_normalization"
                ]
            if "upper_quantile_normalization" in self.config["nucleus_segmentation"]:
                self.normalization_nucleus_segmentation = True
                self.normalization_nucleus_upper_quantile = self.config["nucleus_segmentation"][
                    "upper_quantile_normalization"
                ]

            # check if nuclei should be filtered based on size
            self._check_for_size_filtering(mask_types=["nucleus"])

            # check if nuclei should be filtered based on contact
            if "contact_filter" in self.config["nucleus_segmentation"]:
                self.contact_filter_nuclei = True
                self.contact_filter_nuclei_threshold = self.config["nucleus_segmentation"]["contact_filter"]
            else:
                self.contact_filter_nuclei = False

        if "cytosol_segmentation" in self.config.keys():
            self.segment_cytosol = True
            self.config_cytosol_segmentation = self.config["cytosol_segmentation"]
            # check for additional normalization before segmentation
            # set default values to 0 and 1 so no normalization occurs
            self.normalization_cytosol_lower_quantile = 0
            self.normalization_cytosol_upper_quantile = 1
            self.normalization_cytosol_segmentation = False

            if "lower_quantile_normalization" in self.config["cytosol_segmentation"]:
                self.normalization_cytosol_segmentation = True
                self.normalization_cytosol_lower_quantile = self.config["cytosol_segmentation"][
                    "lower_quantile_normalization"
                ]
            if "upper_quantile_normalization" in self.config_cytosol_segmentation.keys():
                self.normalization_cytosol_segmentation = True
                self.normalization_cytosol_upper_quantile = self.config["cytosol_segmentation"][
                    "upper_quantile_normalization"
                ]

            # check if cytosol should be filtered based on size
            self._check_for_size_filtering(mask_types=["cytosol"])

            # check if cytosol should be filtered based on contact
            if "contact_filter" in self.config["cytosol_segmentation"]:
                self.contact_filter_cytosol = True
                self.contact_filter_cytosol_threshold = self.config["cytosol_segmentation"]["contact_filter"]
            else:
                self.contact_filter_cytosol = False

            # check if cytosol should be filtered based on matching to nuclei
            self._check_for_mask_matching_filtering()

    def _nucleus_segmentation(self, input_image, debug: bool = False):
        if self.normalization_nucleus_segmentation:
            lower = self.normalization_nucleus_lower_quantile
            upper = self.normalization_nucleus_upper_quantile

            self.log(f"Percentile normalization of nucleus input image to range {lower}, {upper}")
            input_image = self._normalize_image(input_image, lower, upper, debug=debug)

        # perform thresholding to generate  a mask of the nuclei

        # Use manual threshold if defined in ["wga_segmentation"]["threshold"]
        # If not, use global otsu
        if "threshold" in self.config["nucleus_segmentation"] and "median_block" in self.config["nucleus_segmentation"]:
            threshold = self.config["nucleus_segmentation"]["threshold"]
            self.log(f"Using local thresholding with a threshold of {threshold} to calculate nucleus mask.")
            self.maps["nucleus_segmentation"] = segment_local_threshold(
                input_image,
                dilation=self.config["nucleus_segmentation"]["dilation"],
                thr=self.config["nucleus_segmentation"]["threshold"],
                median_block=self.config["nucleus_segmentation"]["median_block"],
                min_distance=self.config["nucleus_segmentation"]["min_distance"],
                peak_footprint=self.config["nucleus_segmentation"]["peak_footprint"],
                speckle_kernel=self.config["nucleus_segmentation"]["speckle_kernel"],
                median_step=self.config["nucleus_segmentation"]["median_step"],
                debug=self.debug,
            )
        else:
            self.log("Using global otsu to calculate threshold for nucleus mask generation.")
            self.maps["nucleus_segmentation"] = segment_global_threshold(
                input_image,
                dilation=self.config["nucleus_segmentation"]["dilation"],
                min_distance=self.config["nucleus_segmentation"]["min_distance"],
                peak_footprint=self.config["nucleus_segmentation"]["peak_footprint"],
                speckle_kernel=self.config["nucleus_segmentation"]["speckle_kernel"],
                debug=self.debug,
            )

        # save the thresholding approach as a mask
        self.maps["nucleus_mask"] = np.clip(self.maps["nucleus_segmentation"], 0, 1)

        # calculate nucleus centroids
        # important to do this before any filtering! Otherwise we wont know where to start looking for cytosols
        centers, _, _ = numba_mask_centroid(self.maps["nucleus_segmentation"])
        self.nucleus_centers = centers

        # ensure all edge labels are removed
        self.maps["nucleus_segmentation"] = remove_edge_labels(self.maps["nucleus_segmentation"])

        if self.filter_size:
            self.maps["nucleus_segmentation"] = self._perform_size_filtering(
                self.maps["nucleus_segmentation"],
                self.nucleus_thresholds,  # type: ignore
                self.nucleus_confidence_interval,  # type: ignore
                "nucleus",
                debug=self.debug,
                input_image=input_image if self.debug else None,
            )

        if self.contact_filter_nuclei:
            if self.debug:
                n_classes = len(set(np.unique(self.maps["nucleus_segmentation"])) - {0})

            self.maps["nucleus_segmentation"] = contact_filter(
                self.maps["nucleus_segmentation"],
                threshold=self.contact_filter_nuclei_threshold,
                reindex=False,
            )

            if self.debug:
                n_classes_post = len(set(np.unique(self.maps["nucleus_segmentation"])) - {0})
                self.log(f"Filtered out {n_classes - n_classes_post} nuclei due to contact filtering.")

    def _cytosol_segmentation(self, input_image, debug: bool = False):
        if not self.segment_nuclei:
            raise ValueError("Nucleus segmentation must be performed to be able to perform a cytosol segmentation.")

        if self.normalization_cytosol_segmentation:
            lower = self.normalization_cytosol_lower_quantile
            upper = self.normalization_cytosol_upper_quantile

            input_image = self._normalize_image(input_image, lower, upper, debug=debug)

        # perform thresholding to generate a mask of the cytosol

        # Use manual threshold if defined in ["wga_segmentation"]["threshold"]
        # If not, use global otsu
        if "threshold" in self.config["cytosol_segmentation"]:
            cytosol_mask = input_image < self.config["cytosol_segmentation"]["threshold"]
        else:
            self.log("No treshold for cytosol segmentation defined, global otsu will be used.")
            cytosol_mask = input_image < global_otsu(input_image)

        self._clear_cache(vars_to_delete=[input_image])

        # remove the nucleus mask from the cytosol mask
        cytosol_mask = cytosol_mask.astype(float)
        cytosol_mask -= self.maps["nucleus_mask"]
        cytosol_mask = np.clip(cytosol_mask, 0, 1)

        # Apply dilation and erosion
        cytosol_mask = dilation(cytosol_mask, footprint=disk(self.config["cytosol_segmentation"]["erosion"]))
        cytosol_mask = binary_erosion(
            cytosol_mask,
            footprint=disk(self.config["cytosol_segmentation"]["dilation"]),
        )

        self.maps["cytosol_mask"] = cytosol_mask

        if self.debug:
            fig = plt.figure(figsize=(10, 10))
            plt.imshow(self.maps["cytosol_mask"], cmap="magma")
            plt.axis("off")
            plt.title("Cytosol Mask")
            fig_path = os.path.join(self.directory, "cytosol_mask.png")
            fig.savefig(fig_path)

        self._clear_cache(vars_to_delete=[cytosol_mask])

        # calculate potential map for cytosol segmentation
        potential_map = (
            self.maps["median_corrected"][1] - np.quantile(self.maps["median_corrected"][1], 0.02)
        ) / np.quantile(self.maps["median_corrected"][1], 0.98)
        potential_map = np.clip(potential_map, 0, 1)

        # subtract nucleus mask from potential map
        potential_map = np.clip(potential_map - self.maps["nucleus_mask"], 0, 1)
        potential_map = 1 - potential_map

        # enhance potential map to generate speedmap
        min_clip = self.config["cytosol_segmentation"]["min_clip"]
        max_clip = self.config["cytosol_segmentation"]["max_clip"]
        potential_map = (np.clip(potential_map, min_clip, max_clip) - min_clip) / (max_clip - min_clip)
        potential_map = (potential_map * 0.9 + 0.1).astype(float)

        self.maps["potential_map"] = potential_map

        if self.debug:
            fig = plt.figure(figsize=(10, 10))
            plt.imshow(self.maps["potential_map"], cmap="magma")
            plt.axis("off")
            plt.title("Potential Map")
            fig_path = os.path.join(self.directory, "potential_map.png")
            fig.savefig(fig_path)

        self.log("Cytosol Potential Mask generated.")
        self._clear_cache(vars_to_delete=[potential_map])

        # perform fast marching and watershed segmentation
        fmm_marker = np.ones_like(self.maps["median_corrected"][0])
        px_center = np.round(self.nucleus_centers).astype(np.uint64)

        for center in px_center:
            fmm_marker[center[0], center[1]] = 0

        fmm_marker = np.ma.MaskedArray(fmm_marker, self.maps["cytosol_mask"])
        travel_time = skfmm_travel_time(fmm_marker, self.maps["potential_map"])

        if not isinstance(travel_time, np.ma.core.MaskedArray):
            raise TypeError(
                "Travel_time for WGA based segmentation returned no MaskedArray. This is most likely due to missing WGA background determination."
            )

        self.maps["travel_time"] = travel_time.filled(fill_value=np.max(travel_time))

        marker = np.zeros_like(self.maps["median_corrected"][1])

        for center in px_center:
            marker[center[0], center[1]] = self.maps["nucleus_segmentation"][center[0], center[1]]

        cytosol_labels = watershed(
            self.maps["travel_time"],
            marker.astype(np.int64),
            mask=(self.maps["cytosol_mask"] == 0).astype(np.int64),
        )

        cytosol_segmentation = np.where(self.maps["cytosol_mask"] > 0.5, 0, cytosol_labels)

        # ensure all edge labels are removed
        cytosol_segmentation = remove_edge_labels(cytosol_segmentation)
        self.maps["cytosol_segmentation"] = cytosol_segmentation

        if self.filter_size:
            self.maps["cytosol_segmentation"] = self._perform_size_filtering(
                self.maps["cytosol_segmentation"],
                self.cytosol_thresholds,  # type: ignore
                self.cytosol_confidence_interval,  # type: ignore
                "cytosol",
                debug=self.debug,
                input_image=input_image if self.debug else None,
            )

        if self.contact_filter_cytosol:
            if self.debug:
                n_classes = len(set(np.unique(self.maps["cytosol_segmentation"])) - {0})

            self.maps["cytosol_segmentation"] = contact_filter(
                self.maps["cytosol_segmentation"],
                threshold=self.contact_filter_cytosol_threshold,
                reindex=False,
            )

            if self.debug:
                n_classes_post = len(set(np.unique(self.maps["cytosol_segmentation"])) - {0})
                self.log(f"Filtered out {n_classes - n_classes_post} cytosols due to contact filtering.")

        unique_cytosol_ids = set(np.unique(self.maps["cytosol_segmentation"])) - {0}

        # remove any ids from nucleus mask that dont have a cytosol mask
        self.maps["nucleus_segmentation"][~np.isin(self.maps["nucleus_segmentation"], list(unique_cytosol_ids))] = 0


class WGASegmentation(_ClassicalSegmentation):
    N_MASKS = 2
    N_INPUT_CHANNELS = 2
    MASK_NAMES = ["nucleus", "cytosol"]
    DEFAULT_NUCLEI_CHANNEL_IDS = [0]
    DEFAULT_CYTOSOL_CHANNEL_IDS = [1]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _finalize_segmentation_results(self):
        segmentation = np.stack([self.maps["nucleus_segmentation"], self.maps["cytosol_segmentation"]]).astype(
            self.DEFAULT_SEGMENTATION_DTYPE
        )

        return segmentation

    def _execute_segmentation(self, input_image):
        total_time_start = timeit.default_timer()
        self._get_processing_parameters()

        # intialize maps for storing intermediate results
        self.maps = {}

        input_image = self._transform_input_image(input_image)

        self.maps["input_image"] = input_image

        # normalize input
        if self.normalization_input_image:
            self.maps["normalized"] = self._normalize_image(
                input_image,
                self.lower_quantile_normalization_input_image,
                self.upper_quantile_normalization_input_image,
                debug=self.debug,
            )
            # update input image to normalized image
            input_image = self.maps["normalized"]

        if self.median_filtering_input_image:
            self.maps["median_corrected"] = self._median_correct_image(
                input_image, self.median_filter_size_input_image, debug=self.debug
            )
            # update input image to median corrected image
            input_image = self.maps["median_corrected"]

        start_segmentation = timeit.default_timer()
        if self.segment_nuclei:
            image = input_image[0]
            self._nucleus_segmentation(image, debug=self.debug)

        if self.segment_cytosol:
            image = self.maps["input_image"][1]
            self._cytosol_segmentation(image, debug=self.debug)
        stop_segmentation = timeit.default_timer()
        self.segmentation_time = stop_segmentation - start_segmentation

        if self.debug:
            self._visualize_final_masks()

        all_classes = list(set(np.unique(self.maps["nucleus_segmentation"])) - {0})
        segmentation = self._finalize_segmentation_results()  # type: ignore

        self._save_segmentation_sdata(segmentation, all_classes, masks=self.MASK_NAMES)
        self.total_time = timeit.default_timer() - total_time_start


class ShardedWGASegmentation(ShardedSegmentation):
    method = WGASegmentation

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class DAPISegmentation(_ClassicalSegmentation):
    N_MASKS = 1
    N_INPUT_CHANNELS = 1
    MASK_NAMES = ["nucleus"]
    DEFAULT_NUCLEI_CHANNEL_IDS = [0]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _finalize_segmentation_results(self):
        segmentation = np.stack([self.maps["nucleus_segmentation"]]).astype(self.DEFAULT_SEGMENTATION_DTYPE)

        return segmentation

    def _execute_segmentation(self, input_image):
        total_time_start = timeit.default_timer()
        self._get_processing_parameters()

        # intialize maps for storing intermediate results
        self.maps = {}

        input_image = self._transform_input_image(input_image)
        self.maps["input_image"] = input_image.copy()

        # normalize input
        if self.normalization_input_image:
            self.maps["normalized"] = self._normalize_image(
                input_image,
                self.lower_quantile_normalization_input_image,
                self.upper_quantile_normalization_input_image,
                debug=self.debug,
            )
            # update input image to normalized image
            input_image = self.maps["normalized"]

        if self.median_filtering_input_image:
            self.maps["median_corrected"] = self._median_correct_image(
                input_image, self.median_filter_size_input_image, debug=self.debug
            )
            # update input image to median corrected image
            input_image = self.maps["median_corrected"]

        start_segmentation = timeit.default_timer()
        if self.segment_nuclei:
            self._nucleus_segmentation(input_image[0], debug=self.debug)
        stop_segmentation = timeit.default_timer()
        self.segmentation_time = stop_segmentation - start_segmentation

        all_classes = list(set(np.unique(self.maps["nucleus_segmentation"])) - {0})
        segmentation = self._finalize_segmentation_results()

        self._save_segmentation_sdata(segmentation, all_classes, masks=self.MASK_NAMES)
        self.total_time = timeit.default_timer() - total_time_start


class ShardedDAPISegmentation(ShardedSegmentation):
    method = DAPISegmentation



import numpy as np

from scportrait.pipeline.mask_filtering.filter_segmentation import (
    SegmentationFilter,
    TiledSegmentationFilter,
)
from scportrait.processing.masks.mask_filtering import MatchNucleusCytosolIds


class BaseFiltering(SegmentationFilter):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def get_unique_ids(self, mask):
        return np.unique(mask)[1:]

    def return_empty_mask(self, input_image):
        # write out an empty entry
        self.save_classes(classes={})


class filtering_match_nucleus_to_cytosol(BaseFiltering):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self.filtering_threshold = self.config["filtering_threshold"]

        # allow for optional downsampling to improve computation time
        if "downsampling_factor" in self.config.keys():
            self.N = self.config["downsampling_factor"]
            self.kernel_size = self.config["downsampling_smoothing_kernel_size"]
            self.erosion_dilation = self.config["downsampling_erosion_dilation"]
        else:
            self.N = None
            self.kernel_size = None
            self.erosion_dilation = False

    def process(self, input_masks):
        if isinstance(input_masks, str):
            input_masks = self.read_input_masks(input_masks)

        # perform filtering
        filter = MatchNucleusCytosolIds(
            filtering_threshold=self.filtering_threshold,
            downsampling_factor=self.N,
            erosion_dilation=self.erosion_dilation,
            smoothing_kernel_size=self.kernel_size,
        )
        nucleus_cytosol_pairs = filter.generate_lookup_table(input_masks[0], input_masks[1])

        # save results
        self.save_classes(classes=nucleus_cytosol_pairs)


class multithreaded_filtering_match_nucleus_to_cytosol(TiledSegmentationFilter):
    method = filtering_match_nucleus_to_cytosol



"""
stitch
=======

Functions to assemble tiled images into fullscale mosaics.
Uses out-of-memory computation for the assembly of larger than memory image mosaics.
"""

from ._stitch import ParallelStitcher, Stitcher

__all__ = ["Stitcher", "ParallelStitcher"]



from ._utils import add_scalebar
from .h5sc import cell_grid, cell_grid_multi_channel, cell_grid_single_channel
from .sdata import plot_image, plot_labels, plot_segmentation_mask
from .vis import generate_composite

__all__ = [
    "add_scalebar",
    "cell_grid",
    "cell_grid_multi_channel",
    "cell_grid_single_channel",
    "plot_segmentation_mask",
    "plot_image",
    "plot_labels",
    "generate_composite",
]



import multiprocessing
import os
import timeit
from pathlib import PosixPath

import numpy as np
import torch
from cellpose import models
from skfmm import travel_time as skfmm_travel_time
from skimage.morphology import dilation, disk
from skimage.segmentation import watershed

from scportrait.pipeline._utils.segmentation import (
    numba_mask_centroid,
    remove_edge_labels,
)
from scportrait.pipeline.segmentation.segmentation import (
    ShardedSegmentation,
)
from scportrait.pipeline.segmentation.workflows._base_segmentation_workflow import _BaseSegmentation


class _CellposeSegmentation(_BaseSegmentation):
    def _write_cellpose_seg_params_to_file(self, model_type: str, model_name: str) -> None:
        """
        Writes the cellpose segmentation parameters to a file for debugging/logging purposes
        """
        with open(os.path.join(self.directory, f"cellpose_params_{model_type}.txt"), "w") as f:
            f.write("Cellpose Parameters:\n")
            f.write(f"Model: {model_name}\n")
            f.write(f"Model Type: {model_type}\n")
            f.write(f"Diameter: {self.diameter}\n")
            f.write(f"Resample: {self.resample}\n")
            f.write(f"Flow Threshold: {self.flow_threshold}\n")
            f.write(f"Cellprob Threshold: {self.cellprob_threshold}\n")
            f.write(f"Normalize: {self.normalize}\n")
            f.write(f"Rescale: {self.rescale}\n")

    def _read_cellpose_model(self, modeltype: str, name: str, gpu: str, device) -> models.Cellpose:
        """
        Reads cellpose model based on the modeltype and name. Will load to GPU if available as specified in self._use_gpu

        Parameters
        ----------
        modeltype
            either "pretrained" or "custom" depending on the model to load
        name
            name of the model to load

        Returns
        -------
        cellpose model

        """
        if modeltype == "pretrained":
            model = models.Cellpose(model_type=name, gpu=gpu, device=device)
        elif modeltype == "custom":
            model = models.CellposeModel(pretrained_model=name, gpu=gpu, device=device)
        return model

    def _load_model(self, model_type: str, gpu: str, device) -> models.Cellpose:
        """
        Loads cellpose model

        Parameters
        ----------
        model_type
            either "cytosol" or "nucleus" depending on the model to load

        Returns
        -------
        tuple of expected diameter and the cellpose model
        """

        # load correct segmentation model for cytosol
        if "model" in self.config[f"{model_type}_segmentation"].keys():
            model_name = self.config[f"{model_type}_segmentation"]["model"]
            model = self._read_cellpose_model("pretrained", model_name, gpu=gpu, device=device)

        elif "model_path" in self.config[f"{model_type}_segmentation"].keys():
            model_name = self.config[f"{model_type}_segmentation"]["model_path"]
            if isinstance(model_name, PosixPath):
                model_name = str(model_name)
            model = self._read_cellpose_model("custom", model_name, gpu=gpu, device=device)

        # get model parameters from config if not defined use default values
        if "diameter" in self.config[f"{model_type}_segmentation"].keys():
            self.diameter = self.config[f"{model_type}_segmentation"]["diameter"]
        else:
            self.diameter = None

        if "resample" in self.config[f"{model_type}_segmentation"].keys():
            self.resample = self.config[f"{model_type}_segmentation"]["resample"]
        else:
            self.resample = True

        if "flow_threshold" in self.config[f"{model_type}_segmentation"].keys():
            self.flow_threshold = self.config[f"{model_type}_segmentation"]["flow_threshold"]
        else:
            self.flow_threshold = 0.4

        if "cellprob_threshold" in self.config[f"{model_type}_segmentation"].keys():
            self.cellprob_threshold = self.config[f"{model_type}_segmentation"]["cellprob_threshold"]
        else:
            self.cellprob_threshold = 0.0

        if "normalize" in self.config[f"{model_type}_segmentation"].keys():
            self.normalize = self.config[f"{model_type}_segmentation"]["normalize"]
        else:
            self.normalize = True

        if "rescale" in self.config[f"{model_type}_segmentation"].keys():
            self.rescale = self.config[f"{model_type}_segmentation"]["rescale"]
        else:
            self.rescale = None

        self.log(f"Segmenting {model_type} using the following model: {model_name}")

        self._write_cellpose_seg_params_to_file(model_type=model_type, model_name=model_name)

        return model

    def _check_input_image_dtype(self, input_image: np.ndarray):
        if input_image.dtype != self.DEFAULT_IMAGE_DTYPE:
            if isinstance(input_image.dtype, int):
                ValueError(
                    "Default image dtype is no longer int. Cellpose expects int inputs. Please contact developers."
                )
            else:
                ValueError("Image is not of type uint16, cellpose segmentation expects int input images.")

    def _check_gpu_status(self):
        """
        Checks and updates the GPU status.
        If a multi-GPU setup is used, the function checks the current process and returns the GPU id to use for the segmentation.
        If no GPUs are available, the function defaults to CPU.
        """

        # get GPU information if run with workers
        try:
            current = multiprocessing.current_process()
            cpu_name = current.name
            gpu_id_list = current.gpu_id_list
            cpu_id = int(cpu_name[cpu_name.find("-") + 1 :]) - 1

            if cpu_id >= len(gpu_id_list):
                cpu_id = cpu_id % current.n_processes

            # track gpu_id and update GPU status
            self.gpu_id = gpu_id_list[cpu_id]
            self.status = "multi_GPU"

        except (AttributeError, ValueError):
            # default to single GPU
            self.gpu_id = 0
            self.status = "potentially_single_GPU"

        # check if cuda GPU is available
        if torch.cuda.is_available():
            if self.status == "multi_GPU":
                self.use_GPU = f"cuda:{self.gpu_id}"
                self.device = torch.device(self.use_GPU)
                self.nGPUs = torch.cuda.device_count()
            else:
                self.use_GPU = True
                self.device = torch.device(
                    "cuda"
                )  # dont need to specify id, saying cuda will default to the one thats avaialable
                self.nGPUs = 1

        # check if MPS is available
        elif torch.backends.mps.is_available():
            self.use_GPU = True
            self.device = torch.device("mps")
            self.nGPUs = 1

        # default to CPU
        else:
            self.use_GPU = False
            self.device = torch.device("cpu")
            self.nGPUs = 0

        self.log(
            f"GPU Status for segmentation is {self.use_GPU} and will segment using the following device {self.device}."
        )


class DAPISegmentationCellpose(_CellposeSegmentation):
    N_MASKS = 1
    N_INPUT_CHANNELS = 1
    MASK_NAMES = ["nucleus"]
    DEFAULT_NUCLEI_CHANNEL_IDS = [0]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _setup_filtering(self):
        self._check_for_size_filtering(mask_types=self.MASK_NAMES)

    def _finalize_segmentation_results(self, mask: np.ndarray) -> np.ndarray:
        # ensure correct dtype of the maps

        mask = self._check_seg_dtype(mask=mask, mask_name=self.MASK_NAMES[0])

        segmentation = np.stack([mask])

        return segmentation

    def cellpose_segmentation(self, input_image: np.ndarray) -> np.ndarray:
        self._check_gpu_status()
        self._clear_cache()  # ensure we start with an empty cache

        ################################
        ### Perform Nucleus Segmentation
        ################################

        model = self._load_model(model_type="nucleus", gpu=self.use_GPU, device=self.device)

        if self.normalize is False:
            input_image = (input_image - np.min(input_image)) / (
                np.max(input_image) - np.min(input_image)
            )  # min max normalize to 0-1 range as cellpose expects this

        masks = model.eval(
            [input_image],
            rescale=self.rescale,
            normalize=self.normalize,
            diameter=self.diameter,
            flow_threshold=self.flow_threshold,
            cellprob_threshold=self.cellprob_threshold,
            channels=[1, 0],
        )[0]
        masks = np.array(masks)

        # ensure all edge classes are removed
        masks = remove_edge_labels(masks)

        # check if filtering is required
        self._setup_filtering()

        if self.filter_size:
            masks = self._perform_size_filtering(
                mask=masks,
                thresholds=self.nucleus_thresholds,  # type: ignore
                confidence_interval=self.nucleus_confidence_interval,  # type: ignore
                mask_name="nucleus",
                log=True,
                input_image=input_image if self.debug else None,
            )

        masks = masks.reshape(masks.shape[1:])
        return masks

    def _execute_segmentation(self, input_image):
        total_time_start = timeit.default_timer()

        # check that the correct level of input image is used
        input_image = self._transform_input_image(input_image)

        self._check_input_image_dtype(input_image)

        start_segmentation = timeit.default_timer()
        nucleus_mask = self.cellpose_segmentation(input_image)
        stop_segmentation = timeit.default_timer()
        self.segmentation_time = stop_segmentation - start_segmentation

        # finalize classes list
        all_classes = set(np.unique(nucleus_mask)) - {0}

        segmentation = self._finalize_segmentation_results(mask=nucleus_mask)
        self._save_segmentation_sdata(segmentation, all_classes, masks=self.MASK_NAMES)
        self.total_time = timeit.default_timer() - total_time_start


class ShardedDAPISegmentationCellpose(ShardedSegmentation):
    method = DAPISegmentationCellpose


class NuclearExpansionSegmentationCellpose(DAPISegmentationCellpose):
    N_MASKS = 1
    N_INPUT_CHANNELS = 1
    MASK_NAMES = ["nucleus"]
    DEFAULT_NUCLEUS_CHANNEL_IDS = [0]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        self._get_kernel_config()

    def _get_kernel_config(self):
        if "kernel_size" in self.config.keys():
            self.kernel_size = self.config["kernel_size"]
        else:
            self.kernel_size = 20  # default value

        print(self.kernel_size)

    def _expand_nucleus_mask(self, nucleus_mask: np.ndarray, kernel_size: int) -> np.ndarray:
        """
        Expands the nucleus mask by a given kernel size using dilation

        Parameters
        ----------
        nucleus_mask
            nucleus mask to expand
        kernel_size
            size of the kernel to use for dilation

        Returns
        -------
        expanded nucleus mask
        """
        # get centers of segmented nuclei
        nucleus_centers, _, _ids = numba_mask_centroid(nucleus_mask)
        px_centers = np.round(nucleus_centers).astype(np.uint64)

        # convert to binary mask
        binary_mask = (nucleus_mask > 0).astype(np.uint16)

        # expand the mask using dilation with the given kernel size
        expanded_masks = dilation(binary_mask, disk(kernel_size))

        # initialize marker array with cell_ids
        marker = np.zeros_like(nucleus_mask)
        for center in px_centers:
            marker[center[0], center[1]] = nucleus_mask[center[0], center[1]]

        # perform fast marching to get travel times for watershed
        fmm_marker = np.ones_like(expanded_masks)
        for center in px_centers:
            fmm_marker[center[0], center[1]] = 0

        travel_time = skfmm_travel_time(fmm_marker, expanded_masks)

        # use watershed to get expanded cytosol masks
        cytosol_segmentation = watershed(
            travel_time,
            marker.astype(np.int64),
            mask=(expanded_masks).astype(np.int64),
        )

        # remove edge labels
        cytosol_segmentation = remove_edge_labels(cytosol_segmentation)

        return cytosol_segmentation

    def _execute_segmentation(self, input_image):
        total_time_start = timeit.default_timer()

        # check that the correct level of input image is used
        input_image = self._transform_input_image(input_image)

        self._check_input_image_dtype(input_image)

        start_segmentation = timeit.default_timer()
        nucleus_mask = self.cellpose_segmentation(input_image)
        cytosol_mask = self._expand_nucleus_mask(nucleus_mask, kernel_size=self.kernel_size)
        stop_segmentation = timeit.default_timer()
        self.segmentation_time = stop_segmentation - start_segmentation

        # finalize classes list
        all_classes = set(np.unique(cytosol_mask)) - {0}

        segmentation = self._finalize_segmentation_results(mask=cytosol_mask)
        self._save_segmentation_sdata(segmentation, all_classes, masks=self.MASK_NAMES)
        self.total_time = timeit.default_timer() - total_time_start


class ShardedNuclearExpansionSegmentationCellpose(ShardedSegmentation):
    method = NuclearExpansionSegmentationCellpose


class CytosolSegmentationCellpose(_CellposeSegmentation):
    N_MASKS = 2
    N_INPUT_CHANNELS = 2
    MASK_NAMES = ["nucleus", "cytosol"]
    DEFAULT_NUCLEI_CHANNEL_IDS = [0]
    DEFAULT_CYTOSOL_CHANNEL_IDS = [1]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _finalize_segmentation_results(self, mask_nucleus: np.ndarray, mask_cytosol: np.ndarray) -> np.ndarray:
        # ensure correct dtype of maps

        mask_nucleus = self._check_seg_dtype(mask=mask_nucleus, mask_name="nucleus")
        mask_cytosol = self._check_seg_dtype(mask=mask_cytosol, mask_name="cytosol")

        segmentation = np.stack([mask_nucleus, mask_cytosol])
        return segmentation

    def _setup_filtering(self):
        self._check_for_size_filtering(mask_types=self.MASK_NAMES)
        self._check_for_mask_matching_filtering()

    def cellpose_segmentation(self, input_image: np.ndarray) -> tuple[np.ndarray, np.ndarray]:
        self._check_gpu_status()
        self._clear_cache()  # ensure we start with an empty cache

        ################################
        ### Perform Nucleus Segmentation
        ################################

        model = self._load_model(model_type="nucleus", gpu=self.use_GPU, device=self.device)

        if self.normalize is False:
            input_image = (input_image - np.min(input_image)) / (
                np.max(input_image) - np.min(input_image)
            )  # min max normalize to 0-1 range as cellpose expects this

        masks_nucleus = model.eval(
            [input_image],
            rescale=self.rescale,
            normalize=self.normalize,
            diameter=self.diameter,
            flow_threshold=self.flow_threshold,
            cellprob_threshold=self.cellprob_threshold,
            channels=[1, 0],
        )[0]
        masks_nucleus = np.array(masks_nucleus)  # convert to array

        # manually delete model and perform gc to free up memory on GPU
        self._clear_cache(vars_to_delete=[model])

        # remove edge labels from masks
        masks_nucleus = remove_edge_labels(masks_nucleus)

        #################################
        #### Perform Cytosol Segmentation
        #################################

        model = self._load_model(model_type="cytosol", gpu=self.use_GPU, device=self.device)

        masks_cytosol = model.eval(
            [input_image],
            rescale=self.rescale,
            normalize=self.normalize,
            diameter=self.diameter,
            flow_threshold=self.flow_threshold,
            cellprob_threshold=self.cellprob_threshold,
            channels=[2, 1],
        )[0]
        masks_cytosol = np.array(masks_cytosol)  # convert to array

        # manually delete model and perform gc to free up memory on GPU
        self._clear_cache(vars_to_delete=[model])

        # remove edge labels from masks
        masks_cytosol = remove_edge_labels(masks_cytosol)

        ######################
        ### Perform Filtering to remove too small/too large masks if applicable
        ######################

        # check if filtering is required
        self._setup_filtering()

        if self.filter_size:
            masks_nucleus = self._perform_size_filtering(
                mask=masks_nucleus,
                thresholds=self.nucleus_thresholds,  # type: ignore
                confidence_interval=self.nucleus_confidence_interval,  # type: ignore
                mask_name="nucleus",
                log=True,
                debug=self.debug,
                input_image=input_image if self.debug else None,
            )

            masks_cytosol = self._perform_size_filtering(
                mask=masks_cytosol,
                thresholds=self.nucleus_thresholds,  # type: ignore
                confidence_interval=self.nucleus_confidence_interval,  # type: ignore
                mask_name="cytosol",
                log=True,
                debug=self.debug,
                input_image=input_image if self.debug else None,
            )

        ######################
        ### Perform Filtering match cytosol and nucleus IDs if applicable
        ######################
        self._setup_filtering()

        if self.filter_match_masks:
            masks_nucleus, masks_cytosol = self._perform_mask_matching_filtering(
                nucleus_mask=masks_nucleus,
                cytosol_mask=masks_cytosol,
                input_image=input_image if self.debug else None,
                filtering_threshold=self.mask_matching_filtering_threshold,
                debug=self.debug,
            )
        else:
            self.log(
                "No filtering performed. Cytosol and Nucleus IDs in the two masks do not match. Before proceeding with extraction an additional filtering step needs to be performed"
            )

        ######################
        ### Cleanup Generated Segmentation masks
        ######################

        masks_nucleus = masks_nucleus.reshape(masks_nucleus.shape[1:])
        masks_cytosol = masks_cytosol.reshape(masks_cytosol.shape[1:])

        return (masks_nucleus, masks_cytosol)

    def _execute_segmentation(self, input_image):
        total_time_start = timeit.default_timer()

        # ensure the correct level is selected for the input image
        input_image = self._transform_input_image(input_image)

        # check image dtype since cellpose expects int input images
        self._check_input_image_dtype(input_image)

        start_segmentation = timeit.default_timer()
        masks_nucleus, masks_cytosol = self.cellpose_segmentation(input_image)
        stop_segmentation = timeit.default_timer()
        self.segmentation_time = stop_segmentation - start_segmentation

        # finalize segmentation classes ensuring that background is removed
        all_classes = set(np.unique(masks_nucleus)) - {0}

        segmentation = self._finalize_segmentation_results(mask_nucleus=masks_nucleus, mask_cytosol=masks_cytosol)
        self._save_segmentation_sdata(segmentation, all_classes, masks=self.MASK_NAMES)

        # clean up memory
        self._clear_cache(vars_to_delete=[segmentation, all_classes])
        self.total_time = timeit.default_timer() - total_time_start


class ShardedCytosolSegmentationCellpose(ShardedSegmentation):
    method = CytosolSegmentationCellpose


class CytosolSegmentationDownsamplingCellpose(CytosolSegmentationCellpose):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _finalize_segmentation_results(self, mask_nucleus: np.ndarray, mask_cytosol: np.ndarray) -> np.ndarray:
        mask_nucleus = self._rescale_downsampled_mask(mask_nucleus, "nucleus_segmentation")
        mask_cytosol = self._rescale_downsampled_mask(mask_cytosol, "cytosol_segmentation")

        mask_nucleus = self._check_seg_dtype(mask=mask_nucleus, mask_name="nucleus")
        mask_cytosol = self._check_seg_dtype(mask=mask_cytosol, mask_name="cytosol")

        # combine masks into one stack
        segmentation = np.stack([mask_nucleus, mask_cytosol])

        return segmentation

    def _execute_segmentation(self, input_image):
        total_time_start = timeit.default_timer()

        # ensure the correct level is selected for the input image
        input_image = self._transform_input_image(input_image)

        # check image dtype since cellpose expects int input images
        self._check_input_image_dtype(input_image)

        # only get the first two channels to save memory consumption
        input_image = input_image[:2, :, :]

        # setup downsampling
        self._get_downsampling_parameters()
        self._calculate_padded_image_size(input_image)

        # downsample the image
        input_image = self._downsample_image(input_image)

        start_segmentation = timeit.default_timer()
        mask_nucleus, mask_cytosol = self.cellpose_segmentation(input_image)
        stop_segmentation = timeit.default_timer()
        self.segmentation_time = stop_segmentation - start_segmentation

        # finalize classes list
        all_classes = set(np.unique(mask_nucleus)) - {0}

        segmentation = self._finalize_segmentation_results(mask_nucleus=mask_nucleus, mask_cytosol=mask_cytosol)

        self._save_segmentation_sdata(segmentation, all_classes, masks=self.MASK_NAMES)
        self._clear_cache(vars_to_delete=[segmentation, all_classes])
        self.total_time = timeit.default_timer() - total_time_start


class ShardedCytosolSegmentationDownsamplingCellpose(ShardedSegmentation):
    method = CytosolSegmentationDownsamplingCellpose


class CytosolOnlySegmentationCellpose(_CellposeSegmentation):
    N_MASKS = 1
    N_INPUT_CHANNELS = 2
    MASK_NAMES = ["cytosol"]
    DEFAULT_CYTOSOL_CHANNEL_IDS = [0, 1]

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _setup_filtering(self):
        self._check_for_size_filtering(mask_types=self.MASK_NAMES)

    def _finalize_segmentation_results(self, cytosol_mask: np.ndarray) -> np.ndarray:
        # ensure correct dtype of maps
        cytosol_mask = self._check_seg_dtype(mask=cytosol_mask, mask_name="cytosol")

        segmentation = np.stack([cytosol_mask])
        return segmentation

    def cellpose_segmentation(self, input_image: np.ndarray) -> np.ndarray:
        self._setup_processing()
        self._clear_cache()

        #####
        ### Perform Cytosol Segmentation
        #####

        model = self._load_model(model_type="cytosol", gpu=self.use_GPU, device=self.device)

        if self.normalize is False:
            input_image = (input_image - np.min(input_image)) / (
                np.max(input_image) - np.min(input_image)
            )  # min max normalize to 0-1 range as cellpose expects this

        masks_cytosol = model.eval(
            [input_image],
            rescale=self.rescale,
            normalize=self.normalize,
            diameter=self.diameter,
            flow_threshold=self.flow_threshold,
            cellprob_threshold=self.cellprob_threshold,
            channels=[2, 1],
        )[0]
        masks_cytosol = np.array(masks_cytosol)  # convert to array

        # manually delete model and perform gc to free up memory on GPU
        self._clear_cache(vars_to_delete=[model])

        # ensure edge classes are removed
        masks_cytosol = remove_edge_labels(masks_cytosol)

        #####
        ### Perform Filtering to remove too small/too large masks if applicable
        #####

        self._setup_filtering()

        if self.filter_size:
            masks_cytosol = self._perform_size_filtering(
                mask=masks_cytosol,
                thresholds=self.nucleus_thresholds,  # type: ignore
                confidence_interval=self.nucleus_confidence_interval,  # type: ignore
                mask_name="cytosol",
                log=True,
                debug=self.debug,
                input_image=input_image if self.debug else None,
            )

        masks_cytosol = masks_cytosol.reshape(masks_cytosol.shape[1:])  # add reshape to match shape to HDF5 shape

        return masks_cytosol

    def _execute_segmentation(self, input_image) -> None:
        total_time_start = timeit.default_timer()

        # transform input image
        input_image = self._transform_input_image(input_image)

        # check image dtype since cellpose expects int input images
        self._check_input_image_dtype(input_image)

        # execute segmentation
        start_segmentation = timeit.default_timer()
        cytosol_mask = self.cellpose_segmentation(input_image)
        stop_segmentation = timeit.default_timer()
        self.segmentation_time = stop_segmentation - start_segmentation

        # get final classes list
        all_classes = set(np.unique(cytosol_mask)) - {0}

        segmentation = self._finalize_segmentation_results(cytosol_mask)
        self._save_segmentation_sdata(segmentation, all_classes, masks=self.MASK_NAMES)

        # clean up memory
        self._clear_cache(vars_to_delete=[segmentation, all_classes])

        self.total_time = timeit.default_timer() - total_time_start

        return None


class ShardedCytosolOnlySegmentationCellpose(ShardedSegmentation):
    method = CytosolOnlySegmentationCellpose


class CytosolOnlySegmentationDownsamplingCellpose(CytosolOnlySegmentationCellpose):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def _finalize_segmentation_results(self, cytosol_mask: np.ndarray) -> np.ndarray:
        cytosol_mask = self._rescale_downsampled_mask(cytosol_mask, "cytosol_segmentation")

        cytosol_mask = self._check_seg_dtype(mask=cytosol_mask, mask_name="cytosol")

        # combine masks into one stack
        segmentation = np.stack([cytosol_mask])

        return segmentation

    def _execute_segmentation(self, input_image) -> None:
        total_time_start = timeit.default_timer()

        # ensure the correct level is selected for the input image
        input_image = self._transform_input_image(input_image)

        # check image dtype since cellpose expects int input images
        self._check_input_image_dtype(input_image)

        # get the relevant channels for processing
        input_image = input_image[:2, :, :]
        self.input_image_size = input_image.shape

        # setup downsampling
        self._get_downsampling_parameters()
        self._calculate_padded_image_size(input_image)

        # downsample the image
        input_image = self._downsample_image(input_image)

        start_segmentation = timeit.default_timer()
        cytosol_mask = self.cellpose_segmentation(input_image)
        stop_segmentation = timeit.default_timer()
        self.segmentation_time = stop_segmentation - start_segmentation

        # currently no implemented filtering steps to remove nuclei outside of specific thresholds
        all_classes = set(np.unique(cytosol_mask)) - {0}

        segmentation = self._finalize_segmentation_results(cytosol_mask=cytosol_mask)  # type: ignore

        self._save_segmentation_sdata(segmentation, all_classes, masks=self.MASK_NAMES)
        self.total_time = timeit.default_timer() - total_time_start


class ShardedCytosolOnlySegmentationDownsamplingCellpose(ShardedSegmentation):
    method = CytosolOnlySegmentationDownsamplingCellpose



import os
import shutil
import sys
from concurrent.futures import ThreadPoolExecutor
from pathlib import Path

import numpy as np
from alphabase.io.tempmmap import (
    create_empty_mmap,
    mmap_array_from_path,
    redefine_temp_location,
)
from tqdm.auto import tqdm

from scportrait.io.daskmmap import dask_array_from_path
from scportrait.processing.images._image_processing import rescale_image
from scportrait.tools.stitch._utils.ashlar_plotting import plot_edge_quality, plot_edge_scatter


class Stitcher:
    """
    Class for stitching of image tiles to assemble a mosaic.
    """

    def __init__(
        self,
        input_dir: str,
        slidename: str,
        outdir: str,
        stitching_channel: str,
        pattern: str,
        overlap: float = 0.1,
        max_shift: float = 30,
        filter_sigma: int = 0,
        do_intensity_rescale: bool | str = True,
        rescale_range: tuple = (1, 99),
        channel_order: list[str] = None,
        reader_type="FilePatternReaderRescale",
        image_dtype=None,
        orientation: dict | None = None,
        plot_QC: bool = True,
        overwrite: bool = False,
        cache: str = None,
    ) -> None:
        """Initialize the Stitcher object.

        Args:
            input_dir: Directory containing the input image tiles
            slidename: Name of the slide
            outdir: Output directory to save the stitched mosaic
            stitching_channel: Name of the channel to be used for stitching
            pattern: File pattern to match the image tiles
            overlap: Overlap between adjacent image tiles
            max_shift: Maximum allowed shift during alignment
            filter_sigma: Sigma value for Gaussian filter applied during alignment
            do_intensity_rescale: Flag to rescale image intensities or "full_image" to rescale entire image
            rescale_range: Percentiles for intensity rescaling as tuple or dict with channel names as keys
            channel_order: Order of channels in output mosaic
            reader_type: Type of reader for image tiles
            image_dtype: dtype of the images that are to be stitched, mainly relevant when stitching with the BioformatsReaderRescale
            orientation: Dict specifying dimensions to flip {'flip_x', 'flip_y'}
            plot_QC: Generate quality control figures
            overwrite: Overwrite existing output directory
            cache: Directory for temporary files
        """
        self._lazy_imports()

        if orientation is None:
            orientation = {"flip_x": False, "flip_y": True}

        self.input_dir = input_dir
        self.slidename = slidename
        self.outdir = outdir
        self.stitching_channel = stitching_channel

        # stitching settings
        self.pattern = pattern
        self.overlap = overlap
        self.max_shift = max_shift
        self.filter_sigma = filter_sigma

        # image rescaling
        if do_intensity_rescale == "full_image":
            self.rescale_full_image = True
            self.do_intensity_rescale = True

        else:
            self.do_intensity_rescale = do_intensity_rescale  # type: ignore
            self.rescale_full_image = False
        self.rescale_range = rescale_range

        # setup reader for images
        self.orientation = orientation
        self.reader_type = reader_type
        self.image_dtype = image_dtype

        # workaround for lazy imports of module
        if self.reader_type == "FilePatternReaderRescale":
            self.reader_type = self.FilePatternReaderRescale

        # workflow setup
        self.plot_QC = plot_QC
        self.overwrite = overwrite
        self.channel_order = channel_order
        self.cache = cache

        self._initialize_outdir()

        # initialize variables to default values
        self.reader = None

    def _lazy_imports(self):
        """Import necessary packages for stitching."""
        try:
            import networkx
            import seaborn
            import yattag
            from ashlar import thumbnail
            from ashlar.reg import EdgeAligner, Mosaic
            from ashlar.scripts.ashlar import process_axis_flip
        except ImportError:
            raise ImportError(
                "To use the stitching module, please install the optional stitching dependencies with 'pip install scportrait[stitching]'."
            ) from None

        # check for working java installation
        try:
            from jnius import JavaException, autoclass

        except ImportError:
            raise ImportError(
                "Java is not installed or not configured correctly. Please make sure to install Java e.g. from conda by running 'conda install -c conda-forge openjdk' before trying to stitch data."
            ) from None
        try:
            # Try to access the Java System class
            System = autoclass("java.lang.System")

            # Get Java version
            System.getProperty("java.version")
        except JavaException:
            raise ImportError(
                "Java is not installed or not configured correctly. Please make sure to install Java e.g. from conda by running 'conda install -c conda-forge openjdk' before trying to stitch data."
            ) from None
        from scportrait.tools.stitch._utils.filereaders import (
            BioformatsReaderRescale,
            FilePatternReaderRescale,
        )
        from scportrait.tools.stitch._utils.parallelized_ashlar import ParallelEdgeAligner, ParallelMosaic

        self.ashlar_thumbnail = thumbnail
        self.ashlar_EdgeAligner = EdgeAligner
        self.ashlar_Mosaic = Mosaic
        self.ashlar_process_axis_flip = process_axis_flip
        self.BioformatsReaderRescale = BioformatsReaderRescale
        self.FilePatternReaderRescale = FilePatternReaderRescale
        self.ParallelEdgeAligner = ParallelEdgeAligner
        self.ParallelMosaic = ParallelMosaic

    def __exit__(self):
        self._clear_cache()

    def __del__(self):
        self._clear_cache()

    def _create_cache(self):
        """Create a temporary cache directory for storing intermediate files during stitching."""
        if self.cache is None:
            TEMP_DIR_NAME = redefine_temp_location(self.outdir)
        else:
            TEMP_DIR_NAME = redefine_temp_location(self.cache)

        self.TEMP_DIR_NAME = TEMP_DIR_NAME

    def _clear_cache(self):
        """Clear the temporary cache directory."""
        if "TEMP_DIR_NAME" in self.__dict__:
            if os.path.exists(self.TEMP_DIR_NAME):
                shutil.rmtree(self.TEMP_DIR_NAME)

    def _initialize_outdir(self):
        """Initialize the output directory for saving the stitched mosaic."""
        if not os.path.exists(self.outdir):
            os.makedirs(self.outdir)
            print("Output directory created at: ", self.outdir)
        else:
            if self.overwrite:
                print(f"Output directory at {self.outdir} already exists, overwriting.")
                shutil.rmtree(self.outdir)
                os.makedirs(self.outdir)
            else:
                raise FileExistsError(
                    f"Output directory at {self.outdir} already exists. Set overwrite to True to overwrite the directory."
                )

    def _get_channel_info(self):
        """Get information about the channels in the image tiles."""
        # get channel names
        self.channel_lookup = self.reader.metadata.channel_map
        self.channel_names = list(self.reader.metadata.channel_map.values())
        self.channels = list(self.reader.metadata.channel_map.keys())
        self.stitching_channel_id = list(self.channel_lookup.values()).index(self.stitching_channel)
        self.n_channels = len(self.channels)

    def get_stitching_information(self):
        """Print information about the configuration of the stitching process."""
        if self.reader is None:
            self._initialize_reader()
        self._get_channel_info()

        print("Tile positions will be calculated based on channel:", self.stitching_channel)
        print("Channel Names:", self.channel_names)
        print("Overlap of image tiles:", self.overlap)
        print("Max Shift value:", self.max_shift)
        print("Filter Sigma value:", self.filter_sigma)
        print("Output will be written to:", self.outdir)

    def _setup_rescaling(self):
        """Setup image rescaling based on the specified rescale_range."""
        # set up rescaling
        if self.do_intensity_rescale:
            self.reader.no_rescale_channel = []

            # if all channels should be rescaled to the same range, initialize dictionary with all channels
            if type(self.rescale_range) is tuple:
                self.rescale_range = {k: self.rescale_range for k in self.channel_names}

            # check if all channels are in dictionary for rescaling
            rescale_channels = list(self.rescale_range.keys())

            # make sure all channels provided in lookup dictionary are in the experiment
            if not set(rescale_channels).issubset(set(self.channel_names)):
                raise ValueError("The rescale_range dictionary contains a channel not found in the experiment.")

            # check if we have any channels missing in the rescale_range dictionary
            missing_channels = set.difference(set(self.channel_names), set(rescale_channels))

            if len(missing_channels) > 0:
                Warning(
                    "The rescale_range dictionary does not contain all channels in the experiment."
                    "This may lead to unexpected results. For the missing channels rescaling will be turned off."
                )

                missing_channels = set.difference(self.channel_names, rescale_channels)
                for missing_channel in missing_channels:
                    self.rescale_range[missing_channel] = (0, 1)

                self.reader.no_rescale_channel = [
                    list(self.channel_names).index(missing_channel) for missing_channel in missing_channels
                ]

            # lookup channel names and match them with channel ids to return a new dict whose keys are the channel ids
            rescale_range_ids = {list(self.channel_names).index(k): v for k, v in self.rescale_range.items()}
            self.reader.do_rescale = True
            self.reader.rescale_range = rescale_range_ids  # update so that the lookup can occur correctly

        else:
            self.reader.no_rescale_channel = []
            self.reader.do_rescale = False
            self.reader.rescale_range = None

    def _reorder_channels(self):
        """Reorder the channels in the mosaic based on the specified channel_order."""
        if self.channel_order is None:
            self.channels = self.channels
        else:
            print("current channel order: ", self.channels)

            channels = []
            for channel in self.channel_order:
                channels.append(self.channel_names.index(channel))
            print("new channel order", channels)

            self.channels = channels

    def _initialize_reader(self):
        """Initialize the reader for reading image tiles."""
        if self.reader_type == self.FilePatternReaderRescale:
            self.reader = self.reader_type(
                self.input_dir,
                self.pattern,
                self.overlap,
                rescale_range=self.rescale_range,
                dtype=self.image_dtype,
            )
        elif self.reader_type == self.BioformatsReaderRescale:
            self.reader = self.reader_type(self.input_dir, rescale_range=self.rescale_range)

        # setup correct orientation of slide (this depends on microscope used to generate the data)
        self.ashlar_process_axis_flip(
            self.reader,
            flip_x=self.orientation["flip_x"],
            flip_y=self.orientation["flip_y"],
        )

        # setup rescaling
        self._get_channel_info()
        self._setup_rescaling()

    def save_positions(self):
        """Save the positions of the aligned image tiles."""
        positions = self.aligner.positions
        np.savetxt(
            os.path.join(self.outdir, self.slidename + "_tile_positions.tsv"),
            positions,
            delimiter="\t",
        )

    def generate_thumbnail(self, scale: float | None = 0.05) -> None:
        """Generate a thumbnail of the stitched mosaic.

        Args:
            scale: Scale factor for the thumbnail.
        """
        self._initialize_reader()
        self.thumbnail = self.ashlar_thumbnail.make_thumbnail(
            self.reader, channel=self.stitching_channel_id, scale=scale
        )

        # rescale thumbnail to 0-1 range
        if type(self.rescale_range) is tuple:
            rescale_range = self.rescale_range  # type: ignore
            rescale = True
        elif type(self.rescale_range) is dict:
            rescale_range = tuple(self.rescale_range[self.stitching_channel])  # type: ignore
            rescale = True
        else:
            if not self.do_intensity_rescale:
                rescale = False  # turn off rescaling

        # rescale generated thumbnail
        if rescale:
            self.thumbnail = rescale_image(self.thumbnail, rescale_range)  # type: ignore

    def _initialize_aligner(self):
        """Initialize the aligner for aligning the image tiles.

        Returns:
            Initialized EdgeAligner object.
        """
        aligner = self.ashlar_EdgeAligner(
            self.reader,
            channel=self.stitching_channel_id,
            filter_sigma=self.filter_sigma,
            verbose=True,
            do_make_thumbnail=False,
            max_shift=self.max_shift,
        )
        return aligner

    def plot_qc(self):
        """Plot quality control (QC) figures for the alignment."""
        plot_edge_scatter(self.aligner, self.outdir)
        plot_edge_quality(self.aligner, self.outdir)

    def _perform_alignment(self):
        """Perform alignment of the image tiles."""
        # intitialize reader for getting individual image tiles
        self._initialize_reader()

        print(f"performing stitching on channel {self.stitching_channel} with id number {self.stitching_channel_id}")
        self.aligner = self._initialize_aligner()
        self.aligner.run()

        if self.plot_QC:
            self.plot_qc()

        if self.save_positions:
            self.save_positions()

        self.aligner.reader._cache = {}  # need to empty cache for some reason

        print("Alignment complete.")

    def _initialize_mosaic(self):
        """Initialize the mosaic object for assembling the image tiles.

        Returns:
            Initialized Mosaic object.
        """
        mosaic = self.ashlar_Mosaic(
            self.aligner,
            self.aligner.mosaic_shape,
            verbose=True,
            channels=self.channels,
        )
        return mosaic

    def _assemble_mosaic(self):
        """Assemble the image tiles into a mosaic."""
        # get dimensions of assembled final mosaic
        x, y = self.mosaic.shape
        shape = (self.n_channels, x, y)
        print(f"assembling mosaic with shape {shape}")

        # initialize tempmmap array to save assemled mosaic to
        # if no cache is specified the tempmmap will be created in the outdir

        self._create_cache()

        # create empty mmap array to store assembled mosaic
        hdf5_path = create_empty_mmap(shape, dtype=np.uint16, tmp_dir_abs_path=self.TEMP_DIR_NAME)
        print(f"created tempmmap array for assembled mosaic at {hdf5_path}")
        self.assembled_mosaic = mmap_array_from_path(hdf5_path)
        self.hdf5_path = hdf5_path  # save variable into self for easier access

        # assemble each of the channels
        for i, channel in tqdm(enumerate(self.channels), total=self.n_channels):
            self.assembled_mosaic[i, :, :] = self.mosaic.assemble_channel(
                channel=channel, out=self.assembled_mosaic[i, :, :]
            )

            if self.rescale_full_image:
                # warning this has not been tested for memory efficiency
                print("Rescaling entire input image to 0-1 range using percentiles specified in rescale_range.")
                self.assembled_mosaic[i, :, :] = rescale_image(
                    self.assembled_mosaic[i, :, :], self.rescale_range[channel]
                )

        # convery to dask array
        self.assembled_mosaic = dask_array_from_path(hdf5_path)

    def _generate_mosaic(self):
        # reorder channels
        self._reorder_channels()

        self.mosaic = self._initialize_mosaic()

        # ensure dtype is set correctly
        self.mosaic.dtype = np.uint16
        self._assemble_mosaic()

    def stitch(self):
        """Generate the stitched mosaic."""
        self._perform_alignment()
        self._generate_mosaic()

    def write_tif(self, export_xml: bool = True) -> None:
        """Write the assembled mosaic as TIFF files.

        Args:
            export_xml: Whether to export an XML file for the TIFF files.
                This XML file is compatible with loading the generated TIFF files into BIAS.

        Returns:
            The assembled mosaic are written to file as TIFF files in the specified output directory.
        """

        from scportrait.tools.stitch._utils.filewriters import write_tif, write_xml

        filenames = []
        for i, channel in enumerate(self.channel_names):
            filename = os.path.join(self.outdir, f"{self.slidename}_{channel}.tif")
            filenames.append(filename)
            write_tif(filename, self.assembled_mosaic[i, :, :])

        if export_xml:
            write_xml(filenames, self.channel_names, self.slidename)

    def write_ome_zarr(
        self,
        downscaling_size: int = 4,
        n_downscaling_layers: int = 4,
        chunk_size: tuple[int, int, int] = (1, 1024, 1024),
    ) -> None:
        """Write the assembled mosaic as an OME-Zarr file.

        Args:
            downscaling_size: Downscaling factor for generating lower resolution layers (default is 4).
            n_downscaling_layers: Number of downscaling layers to generate (default is 4).
            chunk_size: Chunk size for the generated OME-Zarr file (default is (1, 1024, 1024)).
        """
        from scportrait.tools.stitch._utils.filewriters import write_ome_zarr

        filepath = os.path.join(self.outdir, f"{self.slidename}.ome.zarr")

        write_ome_zarr(
            filepath,
            self.assembled_mosaic,
            self.channel_names,
            self.slidename,
            overwrite=self.overwrite,
            downscaling_size=downscaling_size,
            n_downscaling_layers=n_downscaling_layers,
            chunk_size=chunk_size,
        )

    def write_thumbnail(self):
        """Write the generated thumbnail as a TIFF file."""

        from scportrait.tools.stitch._utils.filewriters import write_tif

        # calculate thumbnail if this has not already been done
        if "thumbnail" not in self.__dict__:
            self.generate_thumbnail()

        filename = os.path.join(
            self.outdir,
            self.slidename + "_thumbnail_" + self.stitching_channel + ".tif",
        )
        write_tif(filename, self.thumbnail)

    def write_spatialdata(self, scale_factors: list[int] | None = None) -> None:
        """Write the assembled mosaic as a SpatialData object.

        Args:
            scale_factors: List of scale factors for the generated SpatialData object.
                Defaults to [2, 4, 8]. The scale factors are used to generate downsampled versions of the
                image for faster visualization at lower resolutions.
        """
        from scportrait.tools.stitch._utils.filewriters import write_spatialdata

        if scale_factors is None:
            scale_factors = [2, 4, 8]
        filepath = os.path.join(self.outdir, f"{self.slidename}.spatialdata")

        # create spatialdata object
        write_spatialdata(
            filepath,
            image=self.assembled_mosaic,
            channel_names=self.channels,
            scale_factors=scale_factors,
            overwrite=self.overwrite,
        )


class ParallelStitcher(Stitcher):
    """Class for parallel stitching of image tiles and generating a mosaic. For applicable steps multi-threading is used for faster processing."""

    def __init__(
        self,
        input_dir: str,
        slidename: str,
        outdir: str,
        stitching_channel: str,
        pattern: str,
        overlap: float = 0.1,
        max_shift: float = 30,
        filter_sigma: int = 0,
        do_intensity_rescale: bool | str = True,
        rescale_range: tuple = (1, 99),
        channel_order: list[str] = None,
        reader_type="FilePatternReaderRescale",
        image_dtype=None,
        orientation: dict | None = None,
        plot_QC: bool = True,
        overwrite: bool = False,
        cache: str = None,
        threads: int = 20,
    ) -> None:
        """
        Args:
            input_dir: Directory containing the input image tiles.
            slidename: Name of the slide.
            outdir: Output directory to save the stitched mosaic.
            stitching_channel: Name of the channel to be used for stitching.
            pattern: File pattern to match the image tiles.
            overlap: Overlap between adjacent image tiles (default is 0.1).
            max_shift: Maximum allowed shift during alignment (default is 30).
            filter_sigma: Sigma value for Gaussian filter applied during alignment (default is 0).
            do_intensity_rescale: Flag to indicate whether to rescale image intensities (default is True). Alternatively, set to "full_image" to rescale the entire image.
            rescale_range: If all channels should be rescaled to the same range pass a tuple with the percentiles for rescaling (default is (1, 99)). Alternatively, a dictionary can be passed with the channel names as keys and the percentiles as values if each channel should be rescaled to a different range.
            channel_order: Order of channels in the generated output mosaic. If none (default value) the order of the channels is left unchanged.
            reader_type: Type of reader to use for reading image tiles (default is "FilePatternReaderRescale").
            orientation: Dictionary specifying which dimensions of the slide to flip (default is {'flip_x': False, 'flip_y': True}).
            plot_QC: Flag to indicate whether to plot quality control (QC) figures (default is True).
            overwrite: Flag to indicate whether to overwrite the output directory if it already exists (default is False).
            cache: Directory to store temporary files during stitching (default is None). If set to none this directory will be created in the outdir.
            threads: Number of threads to use for parallel processing (default is 20).
        """
        if orientation is None:
            orientation = {"flip_x": False, "flip_y": True}
        super().__init__(
            input_dir=input_dir,
            slidename=slidename,
            outdir=outdir,
            stitching_channel=stitching_channel,
            pattern=pattern,
            overlap=overlap,
            max_shift=max_shift,
            filter_sigma=filter_sigma,
            do_intensity_rescale=do_intensity_rescale,
            rescale_range=rescale_range,
            channel_order=channel_order,
            reader_type=reader_type,
            image_dtype=image_dtype,
            orientation=orientation,
            plot_QC=plot_QC,
            overwrite=overwrite,
            cache=cache,
        )

        # dirty fix to avoide multithreading error with BioformatsReader until this can be fixed
        if self.reader_type == self.BioformatsReaderRescale:
            threads = 1
            print(
                "BioformatsReaderRescale does not support multithreading for calculating the error threshold currently. Proceeding with 1 thread."
            )
            Warning(
                "BioformatsReaderRescale does not support multithreading for calculating the error threshold currently. Proceeding with 1 thread."
            )

        self.threads = threads

    def _initialize_aligner(self):
        """Initialize the aligner for aligning the image tiles.

        Returns:
            aligner (ParallelEdgeAligner): Initialized ParallelEdgeAligner object.
        """
        aligner = self.ParallelEdgeAligner(
            self.reader,
            channel=self.stitching_channel_id,
            filter_sigma=self.filter_sigma,
            verbose=True,
            do_make_thumbnail=False,
            max_shift=self.max_shift,
            n_threads=self.threads,
        )
        return aligner

    def _initialize_mosaic(self):
        mosaic = self.ParallelMosaic(
            self.aligner, self.aligner.mosaic_shape, verbose=True, channels=self.channels, n_threads=self.threads
        )
        return mosaic

    def _assemble_channel(self, args):
        hdf5_path = self.hdf5_path
        channel, i, hdf5_path = args
        out = mmap_array_from_path(hdf5_path)
        self.mosaic.assemble_channel_parallel(channel=channel, ch_index=i, hdf5_path=hdf5_path)

        if self.rescale_full_image:
            # warning this has not been tested for memory efficiency
            print("Rescaling entire input image to 0-1 range using percentiles specified in rescale_range.")
            out[i, :, :] = rescale_image(out[i, :, :], self.rescale_range[channel])

    def _assemble_mosaic(self):
        # get dimensions of assembled final mosaic
        x, y = self.mosaic.shape
        shape = (self.n_channels, x, y)

        print(f"assembling mosaic with shape {shape}")

        self._create_cache()

        hdf5_path = create_empty_mmap(shape, dtype=np.uint16, tmp_dir_abs_path=self.TEMP_DIR_NAME)
        print(f"created tempmmap array for assembled mosaic at {hdf5_path}")

        self.assembled_mosaic = mmap_array_from_path(hdf5_path)
        self.hdf5_path = hdf5_path  # save variable to self for easier access

        # assemble each of the channels
        args = []
        for i, channel in enumerate(self.channels):
            args.append((channel, i, hdf5_path))

        # threading over channels is safe as the channels are written to different postions in the hdf5 file and do not interact with one another
        # threading over the writing of a single channel is not safe and leads to inconsistent results
        workers = np.min([self.threads, self.n_channels])
        print(f"assembling channels with {workers} workers")
        with ThreadPoolExecutor(max_workers=workers) as executor:
            list(executor.map(self._assemble_channel, args))

        # conver to dask array
        self.assembled_mosaic = dask_array_from_path(hdf5_path)

    def write_tif_parallel(self, export_xml: bool = True):
        """Parallelized version of the write_tif method to write the assembled mosaic as TIFF files.

        Args:
            export_xml: Whether to export an XML file for the TIFF files.
                This XML file is compatible with loading the generarted TIFF files into BIAS.

        """
        from scportrait.tools.stitch._utils.filewriters import write_tif, write_xml

        filenames = []
        args = []
        for i, channel in enumerate(self.channel_names):
            filename = os.path.join(self.outdir, f"{self.slidename}_{channel}.tif")
            filenames.append(filename)
            args.append((filename, i))

        tqdm_args = {
            "file": sys.stdout,
            "desc": "writing tif files",
            "total": len(self.channels),
        }

        # define helper function to execute in threadpooler
        def _write_tif(args):
            filename, ix = args
            write_tif(filename, self.assembled_mosaic[ix, :, :])

        # threading over channels is safe as the channels are written to different files
        workers = np.min([self.threads, self.n_channels])
        with ThreadPoolExecutor(max_workers=workers) as executor:
            list(tqdm(executor.map(_write_tif, args), **tqdm_args))

        # write_tif(filename, self.assembled_mosaic[i, :, :])

        if export_xml:
            write_xml(filenames, self.channel_names, self.slidename)



import warnings
from typing import Any, Literal, TypeAlias

import numpy as np
from dask.array.core import Array as daArray
from spatialdata import SpatialData
from spatialdata.models import Image2DModel, Labels2DModel, PointsModel, ShapesModel, TableModel
from spatialdata.transformations.transformations import Identity
from xarray import DataArray, DataTree

from scportrait.tools.sdata.write._helper import add_element_sdata

ChunkSize2D: TypeAlias = tuple[int, int]
ChunkSize3D: TypeAlias = tuple[int, int, int]
ObjectType: TypeAlias = Literal["images", "labels", "points", "tables", "shapes"]

from scportrait.pipeline._utils.constants import DEFAULT_CHUNK_SIZE_2D, DEFAULT_CHUNK_SIZE_3D, DEFAULT_SCALE_FACTORS


def image(
    sdata: SpatialData,
    image: np.ndarray | DataArray | DataTree | daArray,
    image_name: str,
    channel_names: list[str] = None,
    scale_factors: list[int] = None,
    chunks: ChunkSize3D | None = None,
    transform: None = None,
    rgb: bool = False,
    overwrite=False,
) -> None:
    """
    Add the image to the spatialdata object

    Args:
        sdata : SpatialData object to which the image will be added.
        image : Image to be written to the spatialdata object. Can be an array (numpy or dask) or an already validated image model.
        image_name: Name of the image to be written to the spatialdata object.
        channel_names: List of channel names for the image. Default is None.
        scale_factors: List of scale factors for the image. Default is [2, 4, 8]. This will load the image at 4 different resolutions to allow for fluid visualization.
        chunks: Chunk size for the image. Default is (1, 1000, 1000).
        transform: Transformation to be applied to the image. Uses the Identity transformation by default.
        rgb: Whether the image is RGB. Default is False.
        overwrite: Whether to overwrite existing data. Default is False.

    Returns:
        None The spatialdata object is updated on disk.
    """
    # check if the image is already a multi-scale image
    if isinstance(image, DataTree):
        # if so only validate the model since this means we are getting the image from a spatialdata object already
        # fix until #https://github.com/scverse/spatialdata/issues/528 is resolved
        Image2DModel().validate(image)
        if scale_factors is not None:
            warnings.warn("Scale factors are ignored when passing a multi-scale image.", stacklevel=2)
    else:
        if scale_factors is None:
            scale_factors = DEFAULT_SCALE_FACTORS

        if chunks is None:
            chunks = DEFAULT_CHUNK_SIZE_3D

        if rgb:
            dimensions = ["y", "x", "c"]
            channel_names = ["r", "g", "b"]
        else:
            dimensions = ["c", "y", "x"]

        if isinstance(image, DataArray):
            # if so first validate the model since this means we are getting the image from a spatialdata object already
            # fix until #https://github.com/scverse/spatialdata/issues/528 is resolved
            Image2DModel().validate(image)

            if channel_names is not None:
                warnings.warn(
                    "Channel names are ignored when passing a single scale image in the DataArray format. Channel names are read directly from the DataArray.",
                    stacklevel=2,
                )

            if chunks is not None:
                warnings.warn(
                    "Chunks are ignored when passing a single scale image in the DataArray format. Chunks are read directly from the DataArray.",
                    stacklevel=2,
                )

            image = Image2DModel.parse(
                image,
                scale_factors=scale_factors,
                rgb=rgb,
            )

        else:
            if channel_names is None:
                channel_names = [f"channel_{i}" for i in range(image.shape[0])]

            # transform to spatialdata image model
            if transform is None:
                transform_original = Identity()
            else:
                transform_original = transform

            if isinstance(image, daArray):
                # rechunk dask array to match the desired chunk size
                image = image.rechunk(chunks)
                image = Image2DModel.parse(
                    image,
                    dims=dimensions,
                    c_coords=channel_names,
                    scale_factors=scale_factors,
                    transformations={"global": transform_original},
                    rgb=rgb,
                )
            else:
                image = Image2DModel.parse(
                    image,
                    dims=dimensions,
                    chunks=chunks,
                    c_coords=channel_names,
                    scale_factors=scale_factors,
                    transformations={"global": transform_original},
                    rgb=rgb,
                )

    add_element_sdata(sdata, image, image_name, overwrite=overwrite)


def labels(
    sdata: SpatialData,
    labels: np.ndarray | DataArray | DataTree | daArray,
    labels_name: str,
    scale_factors: list[int] = None,
    chunks: ChunkSize2D | None = None,
    overwrite=False,
    transform: None = None,
) -> None:
    """Add the labels to the spatialdata object

    Args:
        sdata : SpatialData object to which the labels will be added.
        labels : Labels to be written to the spatialdata object. Can be an array (numpy or dask) or an already validated labels model.
        labels_name: Name of the labels to be written to the spatialdata object.
        scale_factors: List of scale factors for the labels. Default is [2, 4, 8]. This will load the labels at 4 different resolutions to allow for fluid visualization.
        chunks: Chunk size for the labels.Default is (1000, 1000).
        overwrite: Whether to overwrite existing data.
        transform: Transformation to be applied to the labels. Uses the Identity transformation by default.

    Returns:
        None The spatialdata object is updated on disk.
    """

    if isinstance(labels, DataTree):
        # if so only validate the model since this means we are getting the image from a spatialdata object already
        # fix until #https://github.com/scverse/spatialdata/issues/528 is resolved
        Labels2DModel().validate(labels)
        if scale_factors is not None:
            warnings.warn("Scale factors are ignored when passing a multi-scale label layer.", stacklevel=2)
    else:
        if scale_factors is None:
            scale_factors = DEFAULT_SCALE_FACTORS

        if chunks is None:
            chunks = DEFAULT_CHUNK_SIZE_2D

        if isinstance(labels, DataArray):
            # if so first validate the model since this means we are getting the labels from a spatialdata object already
            # fix until #https://github.com/scverse/spatialdata/issues/528 is resolved
            Labels2DModel().validate(labels)

            if chunks is not None:
                warnings.warn(
                    "Chunks are ignored when passing a single scale image in the DataArray format. Chunks are read directly from the DataArray.",
                    stacklevel=2,
                )

            labels = Labels2DModel.parse(
                labels,
                scale_factors=scale_factors,
            )

        else:
            # transform to spatialdata labels model
            if transform is None:
                transform_original = Identity()
            else:
                transform_original = transform

            dimensions = ["y", "x"]

            if isinstance(labels, daArray):
                # rechunk dask array to match the desired chunk size
                labels = labels.rechunk(chunks)
                labels = Labels2DModel.parse(
                    labels,
                    dims=dimensions,
                    scale_factors=scale_factors,
                    transformations={"global": transform_original},
                )
            else:
                labels = Labels2DModel.parse(
                    labels,
                    dims=dimensions,
                    chunks=chunks,
                    scale_factors=scale_factors,
                )

    add_element_sdata(sdata, labels, labels_name, overwrite=overwrite)



from ._cellpose import (
    CytosolOnlySegmentationCellpose,
    CytosolOnlySegmentationDownsamplingCellpose,
    CytosolSegmentationCellpose,
    CytosolSegmentationDownsamplingCellpose,
    DAPISegmentationCellpose,
    NuclearExpansionSegmentationCellpose,
    ShardedCytosolOnlySegmentationCellpose,
    ShardedCytosolOnlySegmentationDownsamplingCellpose,
    ShardedCytosolSegmentationCellpose,
    ShardedCytosolSegmentationDownsamplingCellpose,
    ShardedDAPISegmentationCellpose,
    ShardedNuclearExpansionSegmentationCellpose,
)
from ._wga_segmentation import (
    DAPISegmentation,
    ShardedDAPISegmentation,
    ShardedWGASegmentation,
    WGASegmentation,
)

__all__ = [
    "Segmentation",
    "ShardedSegmentation",
    "WGASegmentation",
    "ShardedWGASegmentation",
    "DAPISegmentation",
    "ShardedDAPISegmentation",
    "DAPISegmentationCellpose",
    "ShardedDAPISegmentationCellpose",
    "CytosolSegmentationCellpose",
    "ShardedCytosolSegmentationCellpose",
    "CytosolOnlySegmentationCellpose",
    "ShardedCytosolOnlySegmentationCellpose",
    "CytosolSegmentationDownsamplingCellpose",
    "ShardedCytosolSegmentationDownsamplingCellpose",
    "CytosolOnlySegmentationDownsamplingCellpose",
    "ShardedCytosolOnlySegmentationDownsamplingCellpose",
]
